"""
Reader for Web of Science field-tagged bibliographic data.

Tethne parses Web of Science field-tagged data into a list of :class:`.Paper` 
objects. This is a two-step process: data are first parsed into a list of 
dictionaries with field-tags as keys, and then each dictionary is converted to a
:class:`.Paper` . :func:`.readers.wos.read` performs both steps in sequence.

One-step Parsing
````````````````

The method :func:`.readers.wos.read` performs both :func:`.readers.wos.parse` 
and :func:`.readers.wos.convert` . This is the preferred (simplest) approach in
most cases.

.. code-block:: python

   >>> papers = rd.wos.read("/Path/to/savedrecs.txt")
   >>> papers[0]
   <tethne.data.Paper instance at 0x101b575a8>

Alternatively, if you have many data files saved in the same directory, you can 
use :func:`.readers.wos.from_dir` :

.. code-block:: python

   >>> papers = rd.wos.parse_from_dir("/Path/to")

Two-step Parsing
````````````````

Use the two-step approach if you need to access fields not included in 
:class:`.Paper`\, or if you wish to perform some intermediate manipulation on
the raw parsed data.

First import the :mod:`.readers.wos` module:

.. code-block:: python

   >>> import tethne.readers as rd

Then parse the WoS data to a list of field-tagged dictionaries using 
:func:`.readers.wos.parse` :

.. code-block:: python

   >>> wos_list = rd.wos.parse("/Path/to/savedrecs.txt")
   >>> wos_list[0].keys()
   ['EM', '', 'CL', 'AB', 'WC', 'GA', 'DI', 'IS', 'DE', 'VL', 'CY', 'AU', 'JI', 
    'AF', 'CR', 'DT', 'TC', 'EP', 'CT', 'PG', 'PU', 'PI', 'RP', 'J9', 'PT', 
    'LA', 'UT', 'PY', 'ID', 'SI', 'PA', 'SO', 'Z9', 'PD', 'TI', 'SC', 'BP', 
    'C1', 'NR', 'RI', 'ER', 'SN']

Convert those field-tagged dictionaries to :class:`.Paper` objects using 
:func:`.readers.wos.convert` :

.. code-block:: python

   >>> papers = rd.wos.convert(wos_list)
   >>> papers[0]
   <tethne.data.Paper instance at 0x101b575a8>

Methods
```````

.. autosummary::

   convert
   from_dir
   parse
   read

"""

import tethne.data as ds
import xml.etree.ElementTree as ET
import tethne.utilities as util
import os
import re
import uuid

# MACRO for printing the 'print' statement values.
# 0 prints nothing in the console.
# 1 prints all print statements in the console.
DEBUG = 0

# general functions


def _create_ayjid(aulast=None, auinit=None, date=None, jtitle=None, **kwargs):
    """
    Convert aulast, auinit, and jtitle into the fuzzy identifier ayjid
    Returns 'Unknown paper' if all id components are missing (None).

    Parameters
    ----------
    Kwargs : dict
        A dictionary of keyword arguments.
    aulast : string
        Author surname.
    auinit: string
        Author initial(s).
    date : string
        Four-digit year.
    jtitle : string
        Title of the journal.

    Returns
    -------
    ayj : string
        Fuzzy identifier ayjid, or 'Unknown paper' if all id components are
        missing (None).

    """
    if aulast is None:
        aulast = ''
    elif isinstance(aulast, list):
        aulast = aulast[0]

    if auinit is None:
        auinit = ''
    elif isinstance(auinit, list):
        auinit = auinit[0]

    if date is None:
        date = ''

    if jtitle is None:
        jtitle = ''

    ayj = aulast + ' ' + auinit + ' ' + str(date) + ' ' + jtitle

    if ayj == '   ':
        ayj = 'Unknown paper'

    return ayj.upper()

def _create_ainstid(aulast=None, auinit=None, addr1=None, \
                   addr2=None, country=None, **kwargs):
    """
    This function is to create an fuzzy identifier ainstid.
    Convert aulast, auinit, and jtitle into the fuzzy identifier ainstid.
    Returns 'Unknown Institution' if all id components are missing (None).

    Parameters
    ----------
    Kwargs : dict
        A dictionary of keyword arguments.
    aulast : string
        Author surname.
    auinit : string
        Author initial(s).
    address1 : string
        Address of the Institution.
    address2 : string
        Address of the Institution.
    country : string
        Country of affiliation

    Returns
    -------
    ainstid : string
        Fuzzy identifier ainstid, or 'Unknown Institution' if all id components
        are missing (None).

    """
    if aulast is None:
        aulast = ''
    elif isinstance(aulast, list):
        aulast = aulast[0]

    if auinit is None:
        auinit = ''
    elif isinstance(auinit, list):
        auinit = auinit[0]

    if addr1 is None:
        addr1 = ''

    if  addr2 is None:
        addr2 = ''
    if  country is None:
        country = ''

    ainstid = aulast + ' ' + auinit + ' ' + addr1 + ' ' + addr2 + ' ' + country

    if ainstid == ' ':
        ainstid = 'Unknown Institution'

    return ainstid


# Web of Science functions
def parse(filepath):
    """
    Parse Web of Science field-tagged data.

    Parameters
    ----------
    filepath : string
        Filepath to the Web of Science plain text file.

    Returns
    -------
    wos_list : list
        A list of dictionaries each associated with a paper from the Web of
        Science with keys from docs/fieldtags.txt as encountered in the file;
        most values associated with keys are strings with special exceptions
        defined by the list_keys and int_keys variables.

    Raises
    ------
    KeyError : Key value which needs to be converted to an 'int' is not present.
    AttributeError :
    IOError : File at filepath not found, not readable, or empty.

    Examples
    --------

    .. code-block:: python

       >>> import tethne.readers as rd
       >>> wos_list = rd.wos.parse("/Path/to/data.txt")

    Notes
    -----
    Unknown keys: RI, OI, Z9

    """
    wos_list = []

    paper_start_key = 'PT'
    paper_end_key = 'ER'
    stop_flag = 0
    

    # Try to read filepath
    line_list = []
    try:
        with open(filepath,'r') as f:
            line_list = f.read().splitlines()
    except IOError: # File does not exist, or couldn't be read.
        raise IOError("File does not exist, or cannot be read.")

    if len(line_list) is 0:
        raise IOError("Unable to read filepath or filepath is empty.")
    # Convert the data in the file to a usable list of dictionaries.
    # Note: first two lines of file are not related to any paper therein.
    last_field_tag = paper_start_key #initialize to something.
    for line in line_list[2:]:
        line = util.strip_non_ascii(line)
        field_tag = line[:2]
        
        if field_tag == ' ':
            pass
        
        if field_tag == paper_start_key:
            # Then prepare for next paper.
            wos_dict = _new_wos_dict()

        if field_tag == paper_end_key:
            # Then add paper to our list.
            wos_list.append(wos_dict)  

        # Handle keys like AU,AF,CR that continue over many lines.
        if field_tag == '  ':
            field_tag = last_field_tag

        # Add value for the key to the wos_dict: the rest of the line.
        try:
            if field_tag in ['AU', 'AF', 'CR', 'C1', 'CA']:
                # These unique fields use the new line delimiter to distinguish
                # their list elements below.
                # The field C1 can be either in multiple lines or in a single
                # line -- It is the address/institutions of the author.
                wos_dict[field_tag] += '\n' + str(line[3:])
            else:
                wos_dict[field_tag] += ' ' + str(line[3:])
        except (KeyError, TypeError, UnboundLocalError):
                wos_dict[field_tag] = str(line[3:])

        last_field_tag = field_tag
    # End line loop.

    # Define keys that should be lists instead of default string.
    list_keys = ['AU', 'AF', 'DE', 'ID', 'CR', 'C1', 'CA']
    delims = {'AU':'\n',
              'AF':'\n',
              'DE':';',
              'ID':';',
              'C1':'\n',
              'CR':'\n',
              'CA':'\n'}

    # And convert the data at those keys into lists.
    for wos_dict in wos_list:
        for key in list_keys:
            delim = delims[key]
            try:
                key_contents = wos_dict[key]
                if delim != '\n':
                    wos_dict[key] = key_contents.split(delim)
                else:
                    wos_dict[key] = key_contents.splitlines()
            except KeyError:
                # One of the keys to be converted to a list didn't exist.
                pass
            except AttributeError:
                # Again a key didn't exist but it belonged to the wos
                # data_struct set of keys; can't split a None.
                pass

    # Similarly convert some data from string to int.
    int_keys = ['PY']
    for wos_dict in wos_list:
        for key in int_keys:
            try:
                wos_dict[key] = int(wos_dict[key])
            except KeyError:
                # One of the keys to be converted to an int didn't exist.
                pass
            except TypeError:
                # Again a key didn't exist but it belonged to the wos
                # data_struct set of keys; can't convert None to an int.
                pass

    return wos_list

def _parse_cr(ref):

    """
    Supports the Web of Science reader by converting the strings found
    at the CR field tag of a record into a minimum :class:`.Paper` instance.

    Parameters
    ----------
    ref : str
        CR field tag data from a plain text Web of Science file.

    Returns
    -------
    paper : :class:`.Paper`
        A :class:`.Paper` instance.

    Raises
    ------
    IndexError
        When input 'ref' has less number of tokens than necessary ones.
    ValueError
        Gets input with mismacthed inputtype. Ex: getting no numbers for a date
        field.

    Notes
    -----
    Needs a sophisticated name parser, would like to use an open source resource
    for this.

    If WoS is missing a field in the middle of the list there are NOT commas
    indicating that; the following example does NOT occur:

        Doe J, ,, Some Journal

    instead

        Doe J, Some Journal

    This threatens the integrity of WoS data; should we address it?

    Another threat: if WoS is unsure of the DOI number there will be multiple
    DOI numbers in a list of form [doi1, doi2, ...], address this?

    """
    paper = ds.Paper()
    #tokens of form: aulast auinit, date, jtitle, volume, spage, doi
    tokens = ref.split(',')
    try:
        #FIXME: needs better name parser
        # Checking for few parsers, in the meantime trying out few things.
        name = tokens[0]
        # Temp Solution for #62809724
        pattern = re.compile(r'\[(.*?)\]')
        match = pattern.search(name)
        if match:
            # remove the [] and make it a proper one.
            name = name[match.start()+1:match.end()-1]
            if DEBUG :
                print  'stripped name: ', name

        name_tokens = name.split(' ')
        if len(name_tokens) < 2:
            # name_tokens.append('None')
            name_tokens.append(' ')

        paper['aulast'] = [name_tokens[0]]
        paper['auinit'] = [''.join(name_tokens[1:]).replace('.','')]

        if DEBUG:
            print "Final Meta Dicts", paper['aulast'], paper['auinit']

        # Temp Solution for #62809724
        if paper['auinit'] == 'None' or paper['aulast'] == 'None' :
            raise ("The Cited References field is not in the expeceted format")

        #strip initial characters based on the field (spaces, 'V', 'DOI')
        paper['date'] = int(tokens[1][1:])
        paper['jtitle'] = tokens[2][1:]
        paper['volume'] = tokens[3][2:]
        paper['spage'] = tokens[4][2:]
        paper['doi'] = tokens[5][5:]
    except IndexError as E:     # ref did not have the full set of tokens
        pass
    except ValueError as E:     # This occurs when the program expects a date
        pass                    #  but gets a string with no numbers. We leave
                                #  the field incomplete because chances are the
                                #  CR string is too sparse to use anyway.

    ayjid = _create_ayjid(paper['aulast'], paper['auinit'],
                         paper['date'], paper['jtitle'])
    paper['ayjid'] = ayjid

    return paper

def _parse_institutions(ref):
    """
    Supports the Web of Science reader by converting the strings found at the C1
    fieldtag of a record into a minimum :class:`.Paper` instance.

    Parameters
    ----------
    ref : str
        'C1' field tag data from a plain text Web of Science file which contains
        Author First and Last names, Institution affiliated, and the
        location/city where they are affiliated to.

    Returns
    -------
    addr_dict : :class:`.Paper`
        A :class:`.Paper` instance.

    Raises
    ------
    IndexError
        When input 'ref' has less number of tokens than necessary ones.

    ValueError
        Gets input with mismacthed inputtype. Ex: getting no numbers for a date
        field.

    Notes
    -----
    Needs to check many test cases to check various input types.

    """
    addr_dict = ds.Paper()
    #tokens of form:
    tokens = ref.split(',')
    print 'tokens inside parse_institutions : \n', tokens
    try:

        name = tokens[0]
        name_tokens = name.split(' ')
        addr_dict['aulast'] = name_tokens[0]
        addr_dict['auinit'] = name_tokens[1]

        #strip initial characters based on the field (spaces, 'V', 'DOI')
        addr_dict['addr2'] = tokens[1][1:]
        addr_dict['addr3'] = tokens[2][1:]
        addr_dict['country'] = tokens[3][2:]

    except IndexError:
        #ref did not have the full set of tokens
        pass
    except ValueError:
        #this occurs when the program expects a date but gets a string with
        #no numbers, we leave the field incomplete because chances are
        #the CR string is too sparse to use anyway
        pass

    auinsid = _create_ayjid(addr_dict['aulast'], addr_dict['auinit'],
                         addr_dict['date'], addr_dict['jtitle'])
    addr_dict['auinsid'] = auinsid
    print 'auinsid', auinsid
    return addr_dict

def convert(wos_data):
    """
    Convert parsed field-tagged data to :class:`.Paper` instances.

    Convert a dictionary or list of dictionaries with keys from the
    Web of Science field tags into a :class:`.Paper` instance or list of
    :class:`.Paper` instances, the standard for Tethne.
    
    Each :class:`.Paper` is tagged with an accession id for this conversion.

    Parameters
    ----------
    wos_data : list
        A list of dictionaries with keys from the WoS field tags.

    Returns
    -------
    wos_meta : list
        A list of :class:`.Paper` instances.

    Examples
    --------

    .. code-block:: python

       >>> import tethne.readers as rd
       >>> wos_list = rd.wos.parse("/Path/to/data.txt")
       >>> papers = rd.wos.convert(wos_list)

    Notes
    -----
    Need to handle author name anomolies (case, blank spaces, etc.) that may
    make the same author appear to be two different authors in Networkx; this is
    important for any graph with authors as nodes.

    """
    
    accession = str(uuid.uuid4())
    
    #create a Paper for each wos_dict and append to this list
    wos_meta = []

    #handle dict inputs by converting to a 1-item list
    if type(wos_data) is dict:
        wos_data = [wos_data]
        #print 'wos data \n' , wos_data


    # Calling the validate function here, before even building wos_meta list
    # [62809724]
    status = _validate(wos_data)
    if not status:
        #raise Error
        pass

    # Define the direct relationships between WoS fieldtags and Paper keys.
    translator = _wos2paper_map()
    
    # Perform the key convertions
    for wos_dict in wos_data:
        paper = ds.Paper()

        #direct translations
        for key in translator.iterkeys():
            paper[translator[key]] = wos_dict[key]
            
        # Group authors ('CA') are treated as personal authors.
        if 'CA' in wos_dict.keys():
            try: wos_dict['AU'] += wos_dict['CA']
            except TypeError: wos_dict['AU'] = wos_dict['CA']
            try: wos_dict['AF'] += wos_dict['CA']
            except KeyError: wos_dict['AF'] = wos_dict['CA']
    
        # more complicated translations
        # FIXME: not robust to all names, organziation authors, etc.
        if wos_dict['AU'] is not None:
            paper['aulast'], paper['auinit'] = _handle_authors(wos_dict)

        #construct ayjid
        ayjid = _create_ayjid(paper['aulast'], paper['auinit'],
                             paper['date'], paper['jtitle'])
        paper['ayjid'] = ayjid

        # Parse author-institution affiliations. #60216226, #57746858.
        if wos_dict['C1'] is not None:
            paper['institutions'] = _handle_author_institutions(wos_dict)

        # Convert CR references into paper format
        if wos_dict['CR'] is not None:
            meta_cr_list = []
            for ref in wos_dict['CR']:
                meta_cr_list.append(_parse_cr(ref))
                #print 'meta_cr_list' , meta_cr_list
            paper['citations'] = meta_cr_list

        paper['accession'] = accession
        
        wos_meta.append(paper)
    # End wos_dict for loop.

    return wos_meta
    
def _handle_authors(wos_dict):

    aulast_list = []
    auinit_list = []
    for name in wos_dict['AU']:
        name_tokens = name.split(',')
        aulast = name_tokens[0].upper().strip()
        try:
            # 1 for 'aulast, aufirst'
            auinit = name_tokens[1][1].upper().strip()
        except IndexError:
            # then no first initial character
            # preserve parallel name lists with empty string
            auinit = ''
        aulast_list.append(aulast)
        auinit_list.append(auinit)
    
    return aulast_list, auinit_list

def _handle_author_institutions(wos_dict):
    pattern = re.compile(r'\[(.*?)\]')
    author_institutions = {}
    for c1_str in wos_dict['C1']:   # One C1 line for each institution.

        match = pattern.search(c1_str)
        if match:   # Explicit author-institution mappings are provided.
            authors = c1_str[match.start()+1:match.end()-1].split('; ')
            institution = c1_str[match.end():].strip().split(', ')
            for author in authors:
                # The A-I mapping (in data) uses the AF representation
                #  of author names. But we use the AU representation
                #  as our mapping key to ensure consistency with older
                #  datasets.
                author_index = wos_dict['AF'].index(author)
                #e.g."WU, ZD"
                author_au = wos_dict['AU'][author_index].upper()
                inst_name = institution[0]

                try:
                    author_institutions[author_au].append(inst_name)
                except KeyError:
                    author_institutions[author_au] = [inst_name]

        else:   # Author-institution mappings are not provided. We
                #  therefore map all authors to all institutions.
            for author_au in wos_dict['AU']:
                institution = c1_str.strip().split(', ')
                inst_name = institution[0]

                # Use sets here to avoid duplicates.
                try:
                    author_institutions[author_au].add(inst_name)
                except KeyError:
                    author_institutions[author_au] = set([inst_name])

    # Convert values back to lists before returning.
    return { k:list(v) for k,v in author_institutions.iteritems() }

def read(datapath):
    """
    Yields a list of :class:`.Paper` instances from a Web of Science data file.

    Parameters
    ----------
    datapath : string
        Filepath to the Web of Science field-tagged data file.

    Returns
    -------
    papers : list
        A list of :class:`.Paper` instances.
        
    Examples
    --------
    
    .. code-block:: python

       >>> import tethne.readers as rd
       >>> papers = rd.wos.read("/Path/to/data.txt")

    """
    # Added Try Except 
    try:
        wl = parse(datapath)
        papers = convert(wl)
    except IOError:
        raise IOError("Invalid path.")

    return papers

# [#60462784]
def from_dir(path):
    """
    Convenience function for generating a list of :class:`.Paper` from a
    directory of Web of Science field-tagged data files.

    Parameters
    ----------
    path : string
        Path to directory of field-tagged data files.

    Returns
    -------
    papers : list
        A list of :class:`.Paper` objects.

    Raises
    ------
    IOError
        Invalid path.
        
    Examples
    --------

    .. code-block:: python

       >>> import tethne.readers as rd
       >>> papers = rd.wos.from_dir("/Path/to/datadir")        

    """

    wos_list = []

    try:
        files = os.listdir(path)
    except IOError:
        raise IOError("Invalid path.")
            
    for f in files:
        if not f.startswith('.'): # Ignore hidden files.
            try:
                wos_list += parse(path + "/" + f)
            except (IOError,UnboundLocalError): # Ignore files that don't 
                pass                            #  contain WoS data.
    papers = convert(wos_list)
    return papers

# [62809724]
def _validate(wos_data):
    """
    Defines the fucntion to check the input data validation.

    Returns
    -------
    bool - True or false
    if the data is in expected format (True)
    if the respective field is not in expected format (False)

    Raises
    ------
    ValueError - according to the severity of the issue,
    whether the wrong format will affect the further processing.
    """
    if DEBUG:
        print wos_data


    # Create a translator dict whose keys are the fields which needs to be
    # validated from the input.
    # Any new field which needs validation in the future
    translator = _new_wos_dict()

    # Now all these input fields needs to be validated as per requirements.

    for wos_dict in wos_data:
        #direct translations
        for key in translator.iterkeys():
            if DEBUG :
                print wos_dict[key]
            # Validate for 'CR' field
            if wos_dict['CR'] is not None:
                for cr in wos_dict['CR']:
                    # check if the CR field is populated correctly
                    pass
            if wos_dict['C1'] is not None:
                for cr in wos_dict['C1']:
                    # check if the C1 field is populated correctly
                    pass

    status = 1
    return status

def _new_query_dict():
    """
    Declares only those keys of the :class:`.Paper`'s metadata that are
    queryable through CrossRef.
    """
    q_dict = {
                'aulast':None,
                'auinit':None,
                'atitle':None,
                'address':None,
                'jtitle':None,
                'volume':None,
                'issue':None,
                'spage':None,
                'epage':None,
                'date':None     }

    return q_dict


def _new_wos_dict():
    """
    Defines the set of field tags that will try to be converted, and intializes
    them to 'None'.

    Returns
    -------
    wos_dict : dict
        A wos_list dictionary with 'None' as default values for all keys.

    """
    wos_dict = {
                    'DI':None,
                    'AU':None,
                    'C1':None,
                    'TI':None,
                    'SO':None,
                    'VL':None,
                    'IS':None,
                    'BP':None,
                    'EP':None,
                    'PY':None,
                    'UT':None,
                    'CR':None,
                    'AB':None   }

    return wos_dict

def _wos2paper_map():
    """
    Defines the direct relationships between the wos_dict and :class:`.Paper`.

    Returns
    -------
    translator : dict
        A 'translator' dictionary.

    """
    translator = {
                    'DI':'doi',
                    'TI':'atitle',
                    'SO':'jtitle',
                    'VL':'volume',
                    'IS':'issue',
                    'BP':'spage',
                    'EP':'epage',
                    'PY':'date',
                    'UT':'wosid',
                    'AB':'abstract'    }

    return translator

#Custom Error Defined
class DataError(Exception):
    pass



import tethne.writers as wr
import unittest
import networkx as nx
import os

class TestNoneGraph(unittest.TestCase):
    """Test case for graphs with no verticies and no edges"""

    def setUp(self):
        #create None graph
        self.G = nx.Graph()
        self.path = "./testout/G_None"

    def test_sif(self):
        #verify files are written
        wr.graph.to_sif(self.G, self.path)
        self.assertTrue(os.path.isfile(self.path + ".sif"))

        #and they are empty
        if os.path.isfile(self.path + ".sif"):
            G_contents = open(self.path + ".sif",
                              "r").read().splitlines()
            self.assertIs(len(G_contents),0)

    def tearDown(self):
        #remove output files from writer testing
        try:
            os.remove("./testout/G_None.sif")
        except OSError:
            pass

class TestEmptyGraph(unittest.TestCase):
    """Test case for graphs with verticies but no edges"""

    def setUp(self):
        #create Empty graph on n verticies
        self.path = "./testout/G_Empty"
        self.n = 5
        node_list = range(self.n)
        self.G = nx.Graph()
        self.G.add_nodes_from(node_list)

    def test_sif(self):
        #verify files are written
        wr.graph.to_sif(self.G, self.path)
        self.assertTrue(os.path.isfile(self.path + ".sif"))

        #and has a line for each node
        if os.path.isfile(self.path + ".sif"):
            G_contents = open(self.path + ".sif",
                              "r").read().splitlines()
            self.assertIs(len(G_contents), self.n)

    def tearDown(self):
        #remove output files from writer testing
        try:
            os.remove(self.path + ".sif")
        except OSError:
            pass


class TestNAttribEmptyGraph(unittest.TestCase):
    """
    Test case for graphs with verticies with attributes but no edges
    """

    def setUp(self):
        #create Empty graph with node attributes on n verticies
        self.path = "./testout/G_NAttribEmpty"
        self.n = 5
        self.attrib_dict = {'test_att1':1, 'test_att2':2}

        node_list = []
        for i in xrange(self.n):
            node_list.append((i, self.attrib_dict))

        self.G = nx.Graph()
        self.G.add_nodes_from(node_list)


    def test_sif(self):
        #verify files are written
        wr.graph.to_sif(self.G, self.path)
        self.assertTrue(os.path.isfile(self.path + ".sif"))
        for key in self.attrib_dict.iterkeys():
            self.assertTrue(os.path.isfile(self.path + "_" + key + 
                                           ".noa"))

        #and have appropriate contents
        if os.path.isfile(self.path + ".sif"):
            G_contents = open(self.path + ".sif", 
                              "r").read().splitlines()
            self.assertIs(len(G_contents), self.n)

        for key in self.attrib_dict.iterkeys():
            if os.path.isfile(self.path + key + ".noa"):
                noa_contents = open(self.path + "_" + key + ".noa",
                                "r").read().splitlines()
                self.assertIs(len(noa_contents), self.n)

    def tearDown(self):
        #remove output files from writer testing
        try:
            os.remove(self.path + ".sif")
            for key in self.attrib_dict.iterkeys():
                os.remove(self.path + "_" +  key + ".noa")
        except OSError:
            pass

if __name__ == '__main__':
    unittest.main()



"""
Methods for network analysis.
"""

import networkx as nx
import tethne.networks as nt

def node_global_closeness_centrality(g, node, normalize=True):
    """
    Calculates the global closeness centrality of a single node in the network.

    Closeness centrality is based on the average shortest path length
    between a focal node and all other nodes in the network. For multi-component
    graphs, conventional closeness centrality metrics fail because it is not
    possible to traverse between a given node and all other nodes in the graph.
    Global closeness centrality is calculated in a way that yields values even
    for multi-component graphs. For an example of how global closeness
    centrality can be used to analyze co-authorship networks, see the blog post
    `here <http://devo-evo.lab.asu.edu/node/459>`_.

    To calculate the global closeness centrality of a single node, try:

    .. code-block:: python

       >>> import tethne.analyze as az
       >>> ngbc = az.node_global_closeness_centrality(BC, 'LEE 1975 EVOLUTION')
       >>> ngbc
       0.154245

    You can calculate the global closeness centrality of all nodes in the
    network using :func:`.global_closeness_centrality` .

    .. code-block:: python

       >>> GBC = az.global_closeness_centrality(BC)
       >>> GBC
       {'a': 0.0, 'c': 0.0, 'b': 0.6666666666666666, 'd': 0.0}

    For connected graphs, this is equivalent to conventional betweenness
    centrality. For disconnected graphs, works around infinite path lengths
    between nodes in different components.

    Parameters
    ----------
    g : networkx.Graph
    node : any
        Identifier of node of interest in g.
    normalize : boolean
        If True, normalizes centrality based on the average shortest path
        length. Default is True.

    Returns
    -------
    c : float
        Global closeness centrality of node.
    """

    c = 0
    try:
        for pl in nx.shortest_path_length(g, node).values():
            if pl != 0:     # Ignore self-loops.
                c += 1/float(pl)
        c = c/len(g)
    except ZeroDivisionError:
        c = 0.

    if normalize:
        ac = 0
        for sg in nx.connected_component_subgraphs(g):
            if len(sg.nodes()) > 1:
                aspl = nx.average_shortest_path_length(sg)
                ac += (1/aspl) * (float(len(sg)) / float(len(g))**2 )
        c = c/ac

    return c

def global_closeness_centrality(g, normalize=True):
    """
    Calculates global closeness centrality for all nodes in the network.

    See :func:`.node_global_closeness_centrality` for more information.

    Parameters
    ----------
    g : networkx.Graph
    normalize : boolean
        If True, normalizes centrality based on the average shortest path
        length. Default is True.

    Returns
    -------
    C : dict
        Dictionary of results, with node identifiers as keys and gcc as values.
    """

    C = {}
    for node in g.nodes():
        C[node] = node_global_closeness_centrality(g, node, normalize=normalize)

    return C



"""
Classes for building a :class:`.GraphCollection` . 

.. autosummary::

   builder
   authorCollectionBuilder
   paperCollectionBuilder

"""

import networkx as nx
import tethne.networks as nt
from tethne.data import GraphCollection
import types

class builder(object):
    """Base class for builders."""

    def __init__(self, D):
        self.D = D

class paperCollectionBuilder(builder):
    """
    Builds a :class:`.GraphCollection` with method in 
    :mod:`tethne.networks.papers` from a :class:`.DataCollection` .
    """

    def build(self, graph_axis, graph_type, **kwargs):
        """
        Generates graphs for each slice along graph_axis in
        :class:`.DataCollection` D.
        
        Other axes in D are treated as attributes.
        
        **Usage**
    
        .. code-block:: python

           >>> import tethne.readers as rd
           >>> data = rd.wos.read("/Path/to/wos/data.txt")
           >>> from tethne.data import DataCollection
           >>> D = DataCollection(data) # Indexed by wosid, by default.
           >>> D.slice('date', 'time_window', window_size=4)
           >>> from tethne.builders import paperCollectionBuilder
           >>> builder = paperCollectionBuilder(D)
           >>> C = builder.build('date', 'bibliographic_coupling', threshold=2)
           >>> C
           <tethne.data.GraphCollection at 0x104ed3550>
           
           """
        
        # TODO: Check to make sure we have the right stuff.
        
        C = GraphCollection()
        
        # Build a Graph for each slice.
        for key, pids in self.D.axes[graph_axis].iteritems():
            data = [ self.D.data[p] for p in pids ]
            kwargs['node_attribs'] = self.D.get_axes()
            kwargs['node_id'] = self.D.index_by
            C[key] = nt.papers.__dict__[graph_type](data, **kwargs)

        return C

class authorCollectionBuilder(builder):
    """
    Builds a :class:`.GraphCollection` with method in 
    :mod:`tethne.networks.authors` from a :class:`.DataCollection` .
    """
    
    def build(self, graph_axis, graph_type, **kwargs):
        """
        Generates graphs for each slice along graph_axis in
        :class:`.DataCollection` D.
        
        Other axes in D are treated as attributes.
        
        **Usage**
    
        .. code-block:: python

           >>> import tethne.readers as rd
           >>> data = rd.wos.read("/Path/to/wos/data.txt")
           >>> from tethne.data import DataCollection
           >>> D = DataCollection(data) # Indexed by wosid, by default.
           >>> D.slice('date', 'time_window', window_size=4)
           >>> from tethne.builders import authorCollectionBuilder
           >>> builder = authorCollectionBuilder(D)
           >>> C = builder.build('date', 'coauthors')
           >>> C
           <tethne.data.GraphCollection at 0x104ed3550>
           
           """
        
        # TODO: Check to make sure we have the right stuff.
        
        C = GraphCollection()
        
        # Build a Graph for each slice.
        for key, pids in self.D.axes[graph_axis].iteritems():
            data = [ self.D.data[p] for p in pids ]
            C[key] = nt.authors.__dict__[graph_type](data, **kwargs)

        return C    


"""
Export networks to structured and unstructured formats, for visualization.

.. autosummary::

   collection
   graph
   matrix
   
"""

import tethne.writers.graph
import tethne.writers.matrix
import tethne.writers.collection



def isfloat(x):
    try:
        a = float(x)
    except ValueError:
        return False
    else:
        return True

def isint(x):
    try:
        a = float(x)
        b = int(a)
    except ValueError:
        return False
    else:
        return a == b

if __name__ == "__main__":
    
    import csv
    import sys
    import os
    import time
    import math
    import pickle
    from optparse import OptionParser, OptionGroup
    import numpy as np
    
    import tethne.readers as rd
    import tethne.networks as nt
    import tethne.analyze as az
    import tethne.writers as wr
    from tethne.data import DataCollection, GraphCollection
    from tethne.builders import authorCollectionBuilder, paperCollectionBuilder
    
    parser = OptionParser()

    # Only one of the following workflow steps will be carried out per 
    #  script execution.
    workflowGroup = OptionGroup(parser, "Workflow Steps",
                            "Choose one workflow step per script execution."  +\
                            " Each workflow step requires a different set of" +\
                            " additional options.")

    workflowGroup.add_option("--read-file",
                             action="store_true", dest="read_f", default=False,
                             help="Read from a single data file. Requires"    +\
                                  " --data-path and --data-format.")
    workflowGroup.add_option("--read-dir",
                             action="store_true", dest="read_d", default=False,
                             help="Read from a directory containing multiple" +\
                                  " data files. Requires --data-path and"     +\
                                  " --data-format.")
    workflowGroup.add_option("--slice",
                             action="store_true", dest="slice", default=False,
                             help="Slice your dataset for comparison along a" +\
                                  " key axis. Requires --slice-axis. If"      +\
                                  " --outpath is set, produces a table with"  +\
                                  " binned paper frequencies in [OUTPATH]/"   +\
                                  "[DATASET_ID]_slices.csv.")
    workflowGroup.add_option("--graph",
                             action="store_true", dest="graph", default=False,
                            help="Generate a graph (or collection of graphs)."+\
                                 " If --outpath is set, produces a table with"+\
                                 " the number of nodes and edges per graph in"+\
                                 " [OUTPATH]/[DATASET_ID]_graphs.csv/")
    workflowGroup.add_option("--analyze",
                             action="store_true", dest="analyze", default=False,
                            help="Analyze a graph (or collection of graphs)." +\
                                 " If --outpath is set, produces a table with"+\
                                 " the mean and variance of the algorithm"    +\
                                 " result for each graph, in [OUTPATH]/"      +\
                                 "[DATASET_ID]_[ALGORITHM]_analysis.csv.")
    workflowGroup.add_option("--write",
                             action="store_true", dest="write", default=False,
                            help="Write a graph (or collection of graphs) to" +\
                                 " a structured format, in [OUTPATH].")
    
    parser.add_option("-I", "--dataset-id", dest="dataset_id",
                      help="Unique ID (required).")
    parser.add_option("-t", "--temp-dir", dest="temp_dir", default="/tmp",
                      help="Directory for storing temporary files.")
    parser.add_option("-O", "--outpath", dest="outpath",
                      help="Path to save workflow output. Some workflow steps"+\
                           " will generate summary statistics or other output.")

    # Options for read workflow step.
    readGroup = OptionGroup(parser, "Options for read workflow step")
    
    readGroup.add_option("-P", "--data-path", dest="datapath",
                      help="Full path to dataset.")
                      
    readGroup.add_option("-F", "--data-format", dest="dataformat",
                      help="Format of input dataset (WOS, DFR).")

    # Options for slice workflow step.
    sliceGroup = OptionGroup(parser, "Options for slice workflow step")
    
    sliceGroup.add_option("-S", "--slice-axis", dest="slice_axis",
                      help="Key along which to slice the dataset. This can be"+\
                           " one of any of the fields listed in the API " + \
                           " documentation: " + \
                           "./doc/api/tethne.html#tethne.data.Paper .")
                           
    sliceGroup.add_option("-M", "--slice-method", dest="slice_method",
               default="time_period",
               help="Method used to slice DataCollection. Available methods:" +\
                    " time_window, time_period. For details, see"             +\
                    " ./doc/api/tethne.html#tethne.data.DataCollection.slice."+\
                    " Default is time_period.")
                    
    sliceGroup.add_option("--slice-window-size", dest="window_size", default=1,
                        help="Size of slice time-window or period, in years." +\
                             " Default is 1.")
                             
    sliceGroup.add_option("--slice-step-size", dest="step_size",
                          help="Amount to advance time-window in each step" +\
                               " (ignored for time-period).")
    
    sliceGroup.add_option("--cumulative",
                          action="store_true", dest="cumulative", default=False,
                          help="If True, the data from each successive slice" +\
                                " includes the data from all preceding slices.")

    # Required arguments for network-building.
    graphReqGroup = OptionGroup(parser, "Required options for graph workflow" +\
                                        " step.")
                                    
    graphReqGroup.add_option("-N", "--node-type", dest="node_type",
                      help="Must be one of: author, paper, term.")
                      
    graphReqGroup.add_option("-T", "--graph-type", dest="graph_type",
                      help="Name of a network-builing method. Can be one of " +\
                           "any of the methods listed in the API " + \
                           "documentation: " + \
                           "./doc/api/tethne.networks.html#module-tethne."+\
                           "networks. e.g. if '-n' is 'author', '-t' could be"+\
                           " 'coauthors'")

    # Optional arguments for network-building, for setting keyword arguments of
    #  methods in tethne.networks.
    graphGroup = OptionGroup(parser, "Optional network-building arguments",
                             "Each network-building method takes some"+\
                             " optional arguments to control its output. See"+\
                             " ./doc/api/tethne.networks.html for "+\
                             " a complete description of each method and"+\
                             " available arguments. If the selected method"+\
                             " does not accept an argument from the list"+\
                             " below, it will be ignored.")

    graphGroup.add_option("--merged",
                          action="store_true", dest="merged", default=False,
                          help="Ignore DataCollection slicing, and build" +\
                               " a single graph from all Papers.")
                             
    graphGroup.add_option("--threshold", dest="threshold",
                          help="Set the 'threshold' argument.")
                          
    graphGroup.add_option("--topn", dest="topn",
                          help="Set the 'topn' argument.")
                          
    graphGroup.add_option("--node-attr", dest="node_attr",
                          help="List of attributes to include for each node." +\
                               " e.g. --node-attr=date,atitle,jtitle")
                               
    graphGroup.add_option("--edge-attr", dest="edge_attr",
                          help="List of attributes to include for each edge." +\
                               " e.g. --edge-attr=ayjid,atitle,date")
                               
    graphGroup.add_option("--node-id", dest="node_id",
                          help="Field to use as node id (for papers graphs)." +\
                               " e.g. --node-id=ayjid")
                               
    graphGroup.add_option("--weighted", dest="weighted", action="store_true",
                          default=False,
                          help="Trigger the 'weighted' argument.")
    
    # Required arguments for graph analysis.
    analyzeGroup = OptionGroup(parser, "Required options for analyze workflow"+\
                                       " step.")
    
    analyzeGroup.add_option("-A", "--algorithm", dest="algorithm",
                            help="Name of a NetworkX graph analysis algorithm.")
                            
    # Required arguments for graph writing.
    writeGroup = OptionGroup(parser, "Required options for write workflow " +\
                                     "step.")
    writeGroup.add_option("-W", "--write-format", dest="write_format",
                          help="Output format for graph(s). If a static graph"+\
                               " format is chosen (e.g. graphml), each slice" +\
                               " in the GraphCollection will result in a"     +\
                               " separate file. Supported writers: (static)"  +\
                               " graphml, gexf; (dynamic) xgmml")
                               

    parser.add_option_group(workflowGroup)
    parser.add_option_group(readGroup)
    parser.add_option_group(sliceGroup)
    parser.add_option_group(graphReqGroup)
    parser.add_option_group(graphGroup)
    parser.add_option_group(analyzeGroup)
    parser.add_option_group(writeGroup)
    
    (options, args) = parser.parse_args()
    
    
    if options.dataset_id is None:
        sys.exit('Must specify --dataset-id')
    
    #############################
    #   Workflow step: Read     #
    #############################
    if options.read_f or options.read_d:
    
        sys.stdout.write("-"*40 + "\n")
        sys.stdout.write("\tWorkflow step: Read\n")
        sys.stdout.write("-"*40 + "\n")
        
        # Must specify path.
        if options.datapath is None:    
            sys.exit('Specify path to data with --data-path')
            
        # Must specify format.
        if options.dataformat is None:      
            sys.exit('Specify data format with --format argument.')
        
        # Must select a supported format.
        if options.dataformat not in ['WOS', 'DFR']:
            sys.exit('Data format must be one of: WOS, DFR.')
        
        sys.stdout.write("Reading {0} data from file {1}..."
                                  .format(options.dataformat, options.datapath))

        start_time = time.time()

        if options.dataformat == 'WOS':
            if options.read_f:
                papers = rd.wos.read(options.datapath)
            if options.read_d:
                papers = rd.wos.from_dir(options.datapath)
        elif options.dataformat == 'DFR':
            if options.read_f:
                papers = rd.dfr.read(options.datapath)
            if options.read_d:
                papers = rd.dfr.from_dir(options.datapath)

        N = len(papers)
        a = papers[0]['accession']

        t = time.time() - start_time

        sys.stdout.write("done.\n")
        sys.stdout.write("Read {0} papers in {1} seconds. Accession: {2}.\n"
                                                               .format(N, t, a))

        sys.stdout.write("Generating a new DataCollection...")
        if options.dataformat == 'WOS':
            index_by = 'wosid'
        elif options.dataformat == 'DFR':
            index_by = 'doi'
        D = DataCollection(papers, index_by=index_by)
        sys.stdout.write("done.\n")

        # Save DataCollection for next workflow step.
        savepath = options.temp_dir + "/" + \
                   options.dataset_id + "_DataCollection.pickle"
        sys.stdout.write("Saving DataCollection to {0}...".format(savepath))
        pickle.dump(D, open(savepath, 'wb'))
        sys.stdout.write("done.\n")
        
    #############################
    #   Workflow step: Slice    #
    #############################
    elif options.slice:
    
        sys.stdout.write("-"*40 + "\n")
        sys.stdout.write("\tWorkflow step: Slice\n")
        sys.stdout.write("-"*40 + "\n")
        
        # Must specify a slice axis.
        if options.slice_axis is None:
            sys.exit('Must specifify a slice axis with --slice-axis.')
        
        # Load DataCollection.
        loadpath = options.temp_dir + "/" + \
                   options.dataset_id + "_DataCollection.pickle"
        sys.stdout.write("Loading DataCollection from {0}...".format(loadpath))
        sys.stdout.flush()
                
        D = pickle.load(open(loadpath, 'rb'))
        sys.stdout.write("done.\n")
        
        # Slice DataCollection
        axes = options.slice_axis.split(',')
        for a in axes:
            sys.stdout.write("Slicing DataCollection by {0}...".format(a))
            sys.stdout.flush()
            
            if a == 'date':
                if options.step_size is not None:
                    step_size = int(options.step_size)
                else:
                    step_size = 1

                if options.window_size is not None:
                    window_size = int(options.window_size)
                else:
                    window_size = 1

                    
                D.slice(a, method=options.slice_method, 
                           window_size=window_size,
                           step_size=step_size,
                           cumulative=options.cumulative )
            else:
                D.slice(a)
            sys.stdout.write("done.\n")
        
        # Generate binned data. For now, only uses first two axes.
        if options.outpath is not None:
            summarypath = options.outpath + "/" + \
                          options.dataset_id + "_sliceDistribution.csv"
            sys.stdout.write("Saving slice distribution to {0}..."
                                                           .format(summarypath))
            sys.stdout.flush()
                                                                       
            with open(summarypath, 'wb') as f:
                writer = csv.writer(f, delimiter=',')
                if len(axes) == 2:  # 2-dimensional slice binning.
                    A_indices = D.axes[axes[0]].keys()
                    B_indices = D.axes[axes[1]].keys()
                    writer.writerow([''] + A_indices)
                    dist = D.distribution_2d(axes[0], axes[1])
                    for i in xrange(dist.shape[1]):
                        writer.writerow([B_indices[i]] + list(dist[:,i]))
                elif len(axes) == 1:    # 1-dimensional slice binning.
                    dist = D.distribution()
                    A_indices = D.axes[axes[0]].keys()
                    for i in xrange(dist.shape[0]):
                        writer.writerow([A_indices[i], dist[i]])
            sys.stdout.write("done.\n")
        
        # Save sliced DataCollection for next workflow step.
        savepath = options.temp_dir + "/" + \
                   options.dataset_id + "_DataCollection_sliced.pickle"
        sys.stdout.write("Saving sliced DataCollection to {0}..."
                                                              .format(savepath))
        sys.stdout.flush()                                                              
        pickle.dump(D, open(savepath, 'wb'))
        sys.stdout.write("done.\n")
        
    #############################
    #   Workflow step: Graph    #
    #############################
    elif options.graph:
    
        sys.stdout.write("-"*40 + "\n")
        sys.stdout.write("\tWorkflow step: Graph\n")
        sys.stdout.write("-"*40 + "\n")    
    
        # Must specifiy node type.
        if options.node_type is None:
            sys.exit('Must specify node type with --node-type.')
        
        # Must specify graph type.
        if options.graph_type is None:
            sys.exit('Must specifiy a graph type with --graph-type.')
        
        if options.merged:
            loadpath = options.temp_dir + "/" + \
                       options.dataset_id + "_DataCollection.pickle"
            qualifier = "without"
        else:
            loadpath = options.temp_dir + "/" + \
                       options.dataset_id + "_DataCollection_sliced.pickle"
            qualifier = "with"
            
        sys.stdout.write("Loading DataCollection {0} slices from {1}..."
                                                   .format(qualifier, loadpath))
        sys.stdout.flush()                                                    
        
        D = pickle.load(open(loadpath, 'rb'))
        sys.stdout.write("done.\n")

        if not options.merged:
            # Uses the first axis only!
            a = D.get_axes()[0]
            sys.stdout.write("Using first slice in DataCollection: {0}.\n"
                                                                     .format(a))
                                                                 
        # Build kwargs.
        kwargs = {}
        if options.threshold is not None:
            kwargs['threshold'] = int(options.threshold)
        if options.topn is not None:
            if isint(options.topn): topn = int(options.topn)
            elif isfloat(options.topn): topn = float(options.topn)
            kwargs['topn'] = topn
        if options.node_attr is not None:
            kwargs['node_attribs'] = options.node_attr.split(',')
        if options.edge_attr is not None:
            kwargs['edge_attribs'] = options.edge_attr.split(',')
        if options.node_id is not None:
            kwargs['node_id'] = options.node_id
        kwargs['weighted'] = options.weighted
        
        # Build the graph(s).
        sys.stdout.write("Building {0} graph using {1} method..."
                                 .format(options.node_type, options.graph_type))
        sys.stdout.flush()

        start_time = time.time()        
        
        if not options.merged:  # Generate GraphCollection from DataCollection.
            if options.node_type == 'paper':
                builder = paperCollectionBuilder(D)
                
            elif options.node_type == 'author':
                builder = authorCollectionBuilder(D)

            C = builder.build(a, options.graph_type, **kwargs)

        else:   # Generate a single Graph from all of the papers in 
                #  DataCollection, D.
            method = nt.__dict__[options.node_type+'s'] \
                       .__dict__[options.graph_type]
            
            G = method(D.papers(), **kwargs)
            C = GraphCollection()
            C.graphs['all'] = G
            
        sys.stdout.write("done in {0} seconds.\n"
                                                .format(time.time()-start_time))
        if options.node_type == 'term':
            sys.exit('Term graphs not yet implemented. Coming soon!\n')
            
        savepath = options.temp_dir + "/" + \
                   options.dataset_id + "_GraphCollection.pickle"
        sys.stdout.write("Saving GraphCollection to {0}...".format(savepath))
        sys.stdout.flush()
        pickle.dump(C, open(savepath, "wb"))
        sys.stdout.write("done.\n")
        
        if options.outpath is not None:
            resultspath = options.outpath + "/" + \
                          options.dataset_id + "_graphs.csv"
            sys.stdout.write("Writing graph summaries to {0}..."
                                                           .format(resultspath))
            sys.stdout.flush()
            
            with open(resultspath, 'wb') as f:
                writer = csv.writer(f, delimiter=',')
                writer.writerow(['index', 'nodes', 'edges'])
                for key,g in C.graphs.iteritems():
                    writer.writerow([key,len(g.nodes()), len(g.edges())])
            sys.stdout.write("done.\n")
        
    #############################
    #   Workflow step: Analyze  #
    #############################            

    elif options.analyze:   # Only supports collection.algorithm, for now.
    
        sys.stdout.write("-"*40 + "\n")
        sys.stdout.write("\tWorkflow step: Analyze\n")
        sys.stdout.write("-"*40 + "\n")    
        
        if options.algorithm is None:
            sys.exit('Must specify algorithm with --algorithm.')
        
        loadpath = options.temp_dir + "/" + \
                   options.dataset_id + "_GraphCollection.pickle"
        sys.stdout.write("Loading GraphCollection from {0}...".format(loadpath))
        sys.stdout.flush()
        C = pickle.load(open(loadpath, 'rb'))
        sys.stdout.write("done.\n")
        
        sys.stdout.write("Analyzing GraphCollection with {0}..."
                                                     .format(options.algorithm))
        sys.stdout.flush()
        
        r = az.collection.algorithm(C, options.algorithm)
        sys.stdout.write("done.\n")
        
        if options.outpath is not None:
            resultspath = options.outpath + "/" + options.dataset_id + "_" + \
                          options.algorithm + "_analysis.csv"
            sys.stdout.write("Writing graph analysis results to {0}..."
                                                           .format(resultspath))
            sys.stdout.flush()
            
            with open(resultspath, 'wb') as f:
                writer = csv.writer(f, delimiter=',')
                writer.writerow(['index', 'mean', 'variance'])
                for key,g in C.graphs.iteritems():
                    m = np.mean([ n[1][options.algorithm] \
                                                  for n in g.nodes(data=True) ])
                    v = np.var([ n[1][options.algorithm] \
                                                  for n in g.nodes(data=True) ])
                    
                    writer.writerow([key,m,v])
            sys.stdout.write("done.\n")

        # Save GraphCollection (overwrites old GC).
        savepath = options.temp_dir + "/" + \
                   options.dataset_id + "_GraphCollection.pickle"
        sys.stdout.write("Saving GraphCollection to {0}...".format(savepath))
        sys.stdout.flush()
        pickle.dump(C, open(savepath, "wb"))
        sys.stdout.write("done.\n")            
            
    #############################
    #   Workflow step: Write    #
    #############################            

    elif options.write:   # Only supports collection.algorithm, for now.
    
        sys.stdout.write("-"*40 + "\n")
        sys.stdout.write("\tWorkflow step: Write\n")
        sys.stdout.write("-"*40 + "\n")

        # Load GraphCollection.
        loadpath = options.temp_dir + "/" + \
                   options.dataset_id + "_GraphCollection.pickle"
        sys.stdout.write("Loading GraphCollection from {0}...".format(loadpath))
        sys.stdout.flush()
        C = pickle.load(open(loadpath, 'rb'))
        sys.stdout.write("done.\n")
                
        # --write-format is required.
        if options.write_format is None:
            sys.exit('Must specify graph output format with --write-format.')
        
        # --outpath is required.
        if options.outpath is None:
            sys.exit('Must specify output path with --outpath.')
        
        
        # Write the graphs!
        sys.stdout.write("Writing graphs to {0} with format {1}..."
                                 .format(options.outpath, options.write_format))
        sys.stdout.flush()
        
        basepath = options.outpath + "/" + options.dataset_id + "_graph_"
                           
        if options.write_format == 'graphml':
            for k,g in C.graphs.iteritems():
                wr.graph.to_graphml(g, basepath+str(k))
        elif options.write_format == 'xgmml':
            wr.collection.to_dxgmml(C, basepath+"dynamic.xgmml")
        sys.stdout.write("done.\n")


"""
Methods for parsing JSTOR Data-for-Research datasets.

.. autosummary::

   ngrams
   read

"""

import tethne.data as dt
import os
import xml.etree.ElementTree as ET
import re
from tethne.utilities import dict_from_node, strip_non_ascii
from nltk.corpus import stopwords
import uuid

def read(datapath):
    """
    Yields :class:`.Paper` s from JSTOR DfR package.

    Each :class:`.Paper` is tagged with an accession id for this
    read/conversion.

    Parameters
    ----------
    filepath : string
        Filepath to unzipped JSTOR DfR folder containing a citations.XML file.

    Returns
    -------
    papers : list
        A list of :class:`.Paper` objects.

    Examples
    --------

    .. code-block:: python

       >>> import tethne.readers as rd
       >>> papers = rd.dfr.read("/Path/to/DfR")
    """

    try:
        with open(datapath + "/citations.XML", 'rb') as f:
            data = f.read().replace('&', '&amp;')
            root = ET.fromstring(data)
    except IOError:
        raise IOError(datapath+"citations.XML not found.")

    accession = str(uuid.uuid4())

    papers = []
    for article in root:
        paper = _handle_paper(article)
        paper['accession'] = accession
        papers.append(paper)

    return papers

def from_dir(path):
    """
    Convenience function for generating a list of :class:`.Paper` from a
    directory of JSTOR DfR datasets.

    Parameters
    ----------
    path : string
        Path to directory containing DfR dataset directories.

    Returns
    -------
    papers : list
        A list of :class:`.Paper` objects.

    Raises
    ------
    IOError
        Invalid path.

    Examples
    --------

    .. code-block:: python

       >>> import tethne.readers as rd
       >>> papers = rd.dfr.from_dir("/Path/to/datadir")

    """

    papers = []

    try:
        files = os.listdir(path)
    except IOError:
        raise IOError("Invalid path.")  # Ignore hidden files.

    for f in files:
        if not f.startswith('.') and os.path.isdir(path + "/" + f):
            try:
                papers += read(path + "/" + f)
            except (IOError, UnboundLocalError):    # Ignore directories that
                pass                                #  don't contain DfR data.

    return papers

def ngrams(datapath, N='bi', ignore_hash=True, apply_stoplist=False):
    """
    Yields N-grams from a JSTOR DfR dataset.

    Parameters
    ----------
    filepath : string
        Filepath to unzipped JSTOR DfR folder containing N-grams (e.g.
        'bigrams').
    N : string
        'bi', 'tri', or 'quad'
    ignore_hash : bool
        If True, will exclude all N-grams that contain the hash '#' character.
    apply_stoplist : bool
        If True, will exclude all N-grams that contain words in the NLTK
        stoplist.

    Returns
    -------
    ngrams : dict
        Keys are paper DOIs, values are lists of (Ngram, frequency) tuples.

    Examples
    --------

    .. code-block:: python

       >>> import tethne.readers as rd
       >>> trigrams = rd.dfr.ngrams("/Path/to/DfR", N='tri')
    """

    gram_path = datapath + "/" + N + "grams"
    ngrams = {}

    for file in os.listdir(gram_path):
        if file.split('.')[-1] == 'XML':
            root = ET.parse(gram_path + "/" + file).getroot()
            doi = root.attrib['id']
            grams = [ (gram.text.strip(), int(gram.attrib['weight']) )
                       for gram in root.findall(N + 'gram') # v Hashes v
                       if not ignore_hash or '#' not in list(gram.text) ]

            if apply_stoplist:
                stoplist = stoplist.words()
                grams_ = []
                for g,c in grams:
                    for w in g.split():
                        if w not in stoplist:
                            grams_.append( (g,c) )
                grams = grams_

            ngrams[doi] = grams

    return ngrams

def _handle_paper(article):
    """
    Yields a :class:`.Paper` from an article ET node.

    Parameters
    ----------
    article : Element
        ElementTree Element 'article'.

    Returns
    -------
    paper : :class:`.Paper`
    """
    paper = dt.Paper()
    pdata = dict_from_node(article)

    # Direct mappings.
    translator = _dfr2paper_map()
    for key, value in translator.iteritems():
        try:
            paper[value] = str(pdata[key]).upper()
        except KeyError:    # Article may not have all keys of interest.
            pass

    # Handle author names.
    paper['aulast'], paper['auinit'] = _handle_authors(pdata['author'])

    # Handle pubdate.
    paper['date'] = _handle_pubdate(pdata['pubdate'])

    # Handle pagerange.
    paper['spage'], paper['epage'] = _handle_pagerange(pdata['pagerange'])

    # Generate ayjid.
    try:
        paper['ayjid'] = _create_ayjid(paper['aulast'][0], paper['auinit'][0], \
                                       paper['date'], paper['jtitle'])
    except IndexError:  # Article may not have authors.
        pass

    return paper

def _handle_pagerange(pagerange):
    """
    Yields start and end pages from DfR pagerange field.

    Parameters
    ----------
    pagerange : str or unicode
        DfR-style pagerange, e.g. "pp. 435-444".

    Returns
    -------
    start : str
        Start page.
    end : str
        End page.
    """

    try:
        pr = re.compile("pp\.\s([0-9]+)\-([0-9]+)")
        start, end = re.findall(pr, pagerange)[0]
    except IndexError:
        start = end = 0

    return str(start), str(end)

def _handle_pubdate(pubdate):
    """
    Yields a date integer from DfR pubdate field.
    """

    return int(pubdate[0:4])

def _handle_authors(authors):
    """
    Yields aulast and auinit lists from value of authors node.

    Parameters
    ----------
    authors : list, str, or unicode
        Value or values of 'author' element in DfR XML.

    Returns
    -------
    aulast : list
        A list of author surnames (string).
    auinit : list
        A list of author first-initials (string).
    """

    aulast = []
    auinit = []
    if type(authors) is list:
        for author in authors:
            author = str(strip_non_ascii(author))
            try:
                l,i = _handle_author(author)
                aulast.append(l)
                auinit.append(i)
            except ValueError:
                pass
    elif type(authors) is str or type(authors) is unicode:
        author = str(strip_non_ascii(authors))
        try:
            l,i = _handle_author(author)
            aulast.append(l)
            auinit.append(i)
        except ValueError:
            pass
    else:
        raise ValueError("authors must be a list or a string")

    return aulast, auinit

def _handle_author(author):
    """
    Yields aulast and auinit from an author's full name.

    Parameters
    ----------
    author : str or unicode
        Author fullname, e.g. "Richard L. Nixon".

    Returns
    -------
    aulast : str
        Author surname.
    auinit : str
        Author first-initial.
    """

    lname = author.split(' ')

    try:
        auinit = lname[0][0]
        final = lname[-1].upper()
        if final in ['JR.', 'III']:
            aulast = lname[-2].upper() + " " + final.strip(".")
        else:
            aulast = final
    except IndexError:
        raise ValueError("malformed author name")

    return str(aulast), str(auinit)

def _dfr2paper_map():
    """
    Defines the direct relationships between DfR article elements and
    :class:`.Paper` fields.

    Returns
    -------
    translator : dict
        A 'translator' dictionary.
    """

    translator = {  'doi': 'doi',
                    'title': 'atitle',
                    'journaltitle': 'jtitle',
                    'volume': 'volume',
                    'issue': 'issue'    }

    return translator

def _create_ayjid(aulast=None, auinit=None, date=None, jtitle=None, **kwargs):
    """
    Convert aulast, auinit, and jtitle into the fuzzy identifier ayjid
    Returns 'Unknown paper' if all id components are missing (None).

    Parameters
    ----------
    Kwargs : dict
        A dictionary of keyword arguments.
    aulast : string
        Author surname.
    auinit: string
        Author initial(s).
    date : string
        Four-digit year.
    jtitle : string
        Title of the journal.

    Returns
    -------
    ayj : string
        Fuzzy identifier ayjid, or 'Unknown paper' if all id components are
        missing (None).

    """
    if aulast is None:
        aulast = ''
    elif isinstance(aulast, list):
        aulast = aulast[0]

    if auinit is None:
        auinit = ''
    elif isinstance(auinit, list):
        auinit = auinit[0]

    if date is None:
        date = ''

    if jtitle is None:
        jtitle = ''

    ayj = aulast + ' ' + auinit + ' ' + str(date) + ' ' + jtitle

    if ayj == '   ':
        ayj = 'Unknown paper'

    return ayj.upper()



"""
Methods for building networks from terms in bibliographic records. This
includes keywords, abstract terms, etc.

.. autosummary::

   keyword_cooccurrence
   topic_coupling
"""

import numpy as np
import networkx as nx

def keyword_cooccurrence(papers, threshold, connected=False, **kwargs):
    """
    Generates a keyword cooccurrence network.

    Parameters
    ----------
    papers : list
        A list of :class:`.Paper` objects.
    threshold : int
        Minimum number of occurrences for a keyword pair to appear in graph.
    connected : bool
        If True, returns only the largest connected component.

    Returns
    -------
    k_coccurrence :  networkx.Graph
        A keyword coccurrence network.

    """

    # Extract keywords from papers.
    keywords = {}
    for entry in papers:
        if 'keywords' in entry.keys():
            keywords[entry['wosid']] = entry['keywords']

    # Generate the complete set of keywords in the dataset.
    wordset = set([])
    for entry in papers:
        try:
            for kw in keywords[entry['wosid']]:
                wordset.add(kw)
        except:
            pass

    # Mapping of integer indices to keywords.
    i = 0
    dictionary = {}
    dictionary_ = {}
    for word in wordset:
        dictionary[word] = i
        dictionary_[i] = word
        i += 1

    cooccurrence = np.zeros((len(wordset), len(wordset)))
    frequencies = np.zeros((len(wordset),))

    for entry in papers:
        if entry['keywords'] in keywords.keys():
            for word in keywords[entry['wosid']]:
                frequencies[dictionary[word]] += 1
                for word_ in keywords[entry['wosid']]:
                    i = dictionary[word]
                    j = dictionary[word_]
                    if i != j:
                        cooccurrence[i, j] += 1

    G = nx.Graph()
    for i in xrange(len(wordset)):
        for j in xrange(i, len(wordset)):
            if cooccurrence[i, j] > 1 and i != j:
                G.add_edge(dictionary_[i], dictionary_[j], weight=int(cooccurrence[i, j]))

    if connected:   # Return only the first connected component.
        return nx.connected_component_subgraphs(G)[0]
    else:
        return G    # Return the whole graph.

def topic_coupling(model, threshold=0.005, **kwargs):
    """
    Creates a network of words connected by implication in a common topic(s).

    Parameters
    ----------
    model : :class:`.LDAModel`
    threshold : float
        Minimum P(W|T) for coupling.

    Returns
    -------
    tc : networkx.Graph
        A topic-coupling graph, where nodes are terms.
    """

    Z = model.top_word.shape[0]
    W = model.top_word.shape[1]

    edges = {}
    for z in xrange(Z):
        word_sub = []
        for w in xrange(W):
            if model.top_word[z,w] >= threshold:
                word_sub.append(w)

        for i in xrange(len(word_sub)):
            for j in xrange(i+1, len(word_sub)):
                w_i = word_sub[i]
                w_j = word_sub[j]
                p_i = model.top_word[z,w_i]
                p_j = model.top_word[z,w_j]
                try:
                    edges[(w_i,w_j)].append((z, (p_i+p_j)/2))
                except KeyError:
                    edges[(w_i,w_j)] = [(z, (p_i+p_j)/2)]
    tc = nx.Graph()

    for e, topics in edges.iteritems():
        weight = sum( [ t[1] for t in topics ] ) / Z
        i_id = model.vocabulary[e[0]]
        j_id = model.vocabulary[e[1]]
        tc.add_edge(i_id, j_id, weight=weight, topics=[t[0] for t in topics])

    return tc


"""
Build networks from topics in a topic model.
"""

import networkx as nx

def paper_coupling(model, threshold=0.1):
    """
    """

    D = model.doc_topic.shape[0]
    Z = model.doc_topic.shape[1]

    edges = {}
    for d in xrange(D):
        d_s = model.doc_topic[d,:]
        for i in xrange(Z):
            for j in xrange(i+1, Z):
                if d_s[i] >= threshold and d_s[j] >= threshold:
                    try:
                        edges[(i,j)].append( (d, d_s[i]*d_s[j]/2) )
                    except KeyError:
                        edges[(i,j)] = [(d, d_s[i]*d_s[j]/2)]

    pc = nx.Graph()

    for e, papers in edges.iteritems():
        weight = sum( [p[1] for p in papers] ) / D
        pc.add_edge(e[0], e[1], weight=weight, \
                    papers=[model.metadata[p[0]] for p in papers])

    for t in pc.nodes():
        pc.node[t]['words'] = model.top_keys[t][1]  # Add list of top words.

    return pc

def term_coupling(model, threshold=0.01):
    """
    """

    Z = model.top_word.shape[0]
    W = model.top_word.shape[1]

    edges = {}
    for w in xrange(W):
        t_sub = []

        for z in xrange(Z):
            if model.top_word[z,w] >= threshold:
                t_sub.append(z)

        for i in xrange(len(t_sub)):
            for j in xrange(i+1, len(t_sub)):
                t_i = t_sub[i]
                t_j = t_sub[j]
                p_i = model.top_word[t_i,w]
                p_j = model.top_word[t_j,w]
                try:
                    edges[(t_i,t_j)].append((w, (p_i+p_j)/2))
                except KeyError:
                    edges[(t_i,t_j)] = [(w, (p_i+p_j)/2)]
    tc = nx.Graph()

    #print edges

    for e, words in edges.iteritems():
        weight = sum( [ w[1] for w in words ] ) / W
        word_list = [model.vocabulary[w[0]] for w in words]
        tc.add_edge(e[0], e[1], weight=weight, words=word_list)

    for t in tc.nodes():
        tc.node[t]['words'] = model.top_keys[t][1]  # Add list of top words.

    return tc


import sys
sys.path.append("../tethne")
import tethne.readers as rd
import tethne.data as dt
import unittest

class TestRead(unittest.TestCase):

    def setUp(self):
        filepath = "./testin/dfr"
        self.papers = rd.dfr.read(filepath)

    def test_number(self):
        """
        Each article should be converted.
        """

        self.assertEqual(len(self.papers), 4)

    def test_type(self):
        """
        Should produce a list of :class:`.Paper`
        """

        self.assertIs(type(self.papers[0]), dt.Paper)

    def test_handle_authors(self):
        """
        _handle_authors should generate lists of author surnames (aulast) and
        author first-initialis (auinit) of the same length.
        """

        aulast = self.papers[0]['aulast']
        expected = ['MADGWICK', 'SATOO']
        self.assertListEqual(aulast, expected )

        auinit = self.papers[0]['auinit']
        expected = ['H', 'T']
        self.assertListEqual(auinit, expected )

        self.assertEqual(len(aulast), len(auinit))

    def test_handle_authors_raises(self):
        """
        If parameter is not a list or a string, should raise a ValueError.
        """

        with self.assertRaises(ValueError):
            rd.dfr._handle_authors(1234)

    def test_handle_pubdate(self):
        """
        _handle_pubdate should yield a four-digit integer from a DfR pubdate
        string.
        """

        pubdate = self.papers[0]['date']
        self.assertIs(type(pubdate), int)   # Integer

        self.assertEqual(len(str(pubdate)), 4) # Four-digit

    def test_handle_pagerange(self):
        """
        _handle_pagerange should yield start and end pages, as strings.
        """

        start = self.papers[0]['spage']
        end = self.papers[0]['epage']

        self.assertIs(type(start), str) # Strings
        self.assertIs(type(end), str)

    def teardown(self):
        pass

if __name__ == '__main__':
    unittest.main()


import unittest
import tethne.readers as rd
import tethne.utilities as util
import os
from pprint import pprint


class TestConvert(unittest.TestCase):

    def setUp(self):
       filepath =  "./testin/authorinstitutions_test.txt"
       self.wos_list  = rd.wos.parse(filepath)
       self.meta_list = rd.wos.convert(self.wos_list)


    def test_institutions_1(self):
        # Define expected values for C1 meta_dict key
        # CASES : TESTED
        # Overlapping of authors - whether append works correctly or not
        # KeyError is not creating exception issues.

        institutions = self.meta_list[0]['institutions']

        expected_dict1 =  {'ZHANG, YC': ['Victoria Univ'],\
                            'WU, ZD': \
                                ['Wenzhou Univ', 'Univ Sci & Technol China'],
                            'LU, CL': \
                                ['Wenzhou Univ', 'Northwestern Polytech Univ'],
                            'CHEN, EH': ['Univ Sci & Technol China'],
                            'ZHANG, H': \
                                ['Inst Sci & Technol Informat Zhejiang Prov'],
                            'XU, GD': ['Univ Technol Sydney']}

        self.assertDictEqual(institutions, expected_dict1,"Not equal")

        # Writing the remaining tests.

    def test_institutions_2(self):
        # Define expected values for C1 meta_dict key
        # 2 authors - 2 institutions .. both should be mapped.
        # No overlapping or missing out of expected mappings.

        institutions = self.meta_list[1]['institutions']
        expected_dict2 =  {'VITIELLO, A': ['Univ Salerno'],\
                            'LOIA, V': ['Univ Salerno'], \
                            'ACAMPORA, G': ['Eindhoven Univ Technol']}

        # Check if the expected and resulted 'institutions' field are the same.
        # Changed to Dict Check from List check
        self.assertDictEqual(institutions, expected_dict2,"Not equal")

   # Commented the following block as it will throw Key Error ( as expected).
   # We can uncomment it anytime.
   #============================================================================

    def test_institutions_3(self):
        # define expected values for C1 meta_dict key

        # CASES : TESTED
        # What if the record did not contain C1 input field info.
        institutions = self.meta_list[2]['institutions']
        expected_dict3 = {}
        try:
           institutions = self.meta_list[2]['institutions']
        except AssertationError:
           self.assertDictEqual(institutions, expected_dict3,"Not equal")


    def test_institutions_4(self):
        # define expected values for C1 meta_dict key

        # No authors in the C1 field and the N authors must be mapped with
        # 1 institution - N-1 Mapping

        institutions = self.meta_list[3]['institutions']
        expected_dict4 = {'Huang, TCK': ['Natl Chung Cheng Univ']}

        #Check if the expected and resulted 'institutions' field are the same.
        #self.assertListEqual(institutions, expected_dict4,"Not equal")
        self.assertDictEqual(institutions, expected_dict4,"Not equal")
    

    

# Commenting the Pubmed test cases as of now.
#
#class TestPubmedXmlParse(unittest.TestCase):
#
#    def setUp(self):
#       filepath = './testin/pmc_sample1.xml'
#       self.meta_list = rd.pubmed.parse_pubmed_xml(filepath)
#
#    def test_schema_validation(self):
#        pass
#
#    def test_sample1_front(self):
#        # define expected values for each meta_dict key
#        # citations left out for front matter test
#        expected = {'aulast':['Peng', 'Yuan', 'Wang'],
#                    'auinit':['J', 'J', 'J'],
#                    'atitle':('Effect of Diets Supplemented with Different' +
#                              ' Sources of Astaxanthin on the Gonad of the' +
#                              ' Sea Urchin Anthocidaris crassispina'),
#                    'jtitle':'Nutrients',
#                    'volume':4,
#                    'issue':8,
#                    'spage':922,
#                    'epage':934,
#                    'date':2012,
#                    'ayjid':'Peng J 2012 Nutrients',
#                    'doi':'10.3390/nu4080922',
#                    'pmid':23016124,
#                    'wosid':None}
#
#        obtained = self.meta_list[0]
#        for key in expected.iterkeys():
#            self.assertEqual(expected[key], obtained[key])
#
#    def test_sample1_back(self):
#        pass
#
#    def test_sample2_result(self):
#        pass
#
#    def tearDown(self):
#        pass
#

# Cases for parsing files from a directory.

    def test_from_dir(self):
    
        self.dir_path = "../testsuite/testin/only3"
        self.wrong_path = "../testsuite/testin232/only3"
        
        # Check if objects
        a  = rd.wos.from_dir(self.dir_path)
        self.assertIsInstance(a,type(a), msg="yes it is a list of objects")
        
        ayjid_list = []
        for obj in a :
            ayjid_list.append(obj['ayjid'])
            
                
        # NEED TO FORMAT THESE. 
        expected_ayjid_list = ['MODIS K 2014 ', 'TUERHONG G 2014 ', 'WU Z 2013 NEUROCOMPUTING', 'ACAMPORA G 2013 INFORMATION SCIENCES', 'HRISTOSKOVA A 2013 AUTOMATED SOFTWARE ENGINEERING', 'SUN F 2013 NEUROCOMPUTING', 'WU Z 2013 NEUROCOMPUTING', 'ACAMPORA G 2013 INFORMATION SCIENCES', 'ZADEH P 2013 INFORMATION SCIENCES', 'KEATOR D 2013 NEUROIMAGE', 'POELMANS J 2013 EXPERT SYSTEMS WITH APPLICATIONS', 'SANCHEZ D 2013 INFORMATION SCIENCES', 'DENG J 2013 IEEE TRANSACTIONS ON KNOWLEDGE AND DATA ENGINEERING', 'TAHIR A 2013 JOURNAL OF SYSTEMS AND SOFTWARE', 'ZHONG Y 2013 COMPUTER-AIDED DESIGN', 'HAN L 2013 ADVANCES IN ENGINEERING SOFTWARE', 'RIOS-ALVARADO A 2013 EXPERT SYSTEMS WITH APPLICATIONS', 'DUAN J 2013 EXPERT SYSTEMS WITH APPLICATIONS', 'ABANDA F 2013 EXPERT SYSTEMS WITH APPLICATIONS', 'DUTTA R 2013 IEEE SENSORS JOURNAL', 'KARLA P 2013 JOURNAL OF MEDICAL SYSTEMS', 'CHENG S 2013 WIRELESS PERSONAL COMMUNICATIONS', 'SINACI A 2013 JOURNAL OF BIOMEDICAL INFORMATICS', 'TAO C 2013 JOURNAL OF BIOMEDICAL INFORMATICS', 'BALLATORE A 2013 KNOWLEDGE AND INFORMATION SYSTEMS', 'GOASDOUE F 2013 VLDB JOURNAL', 'FURCHE T 2013 VLDB JOURNAL', 'BOZZON A 2013 VLDB JOURNAL', 'GRACY K 2013 JOURNAL OF THE AMERICAN SOCIETY FOR INFORMATION SCIENCE AND TECHNOLOGY', 'GUNS R 2013 JOURNAL OF THE AMERICAN SOCIETY FOR INFORMATION SCIENCE AND TECHNOLOGY', 'CUXAC P 2013 SCIENTOMETRICS', 'CASES M 2013 JOURNAL OF INTERNAL MEDICINE', 'FELBER P 2013 SOFTWARE-PRACTICE & EXPERIENCE', 'VIZCARRA J 2013 JOURNAL OF WEB ENGINEERING', 'MANGIONE G 2013 JOURNAL OF WEB ENGINEERING', 'ZHANG Z 2013 NATURAL LANGUAGE ENGINEERING', 'ANGROSH M 2013 NATURAL LANGUAGE ENGINEERING', 'CHEN Y 2013 IEEE TRANSACTIONS ON KNOWLEDGE AND DATA ENGINEERING', 'LAZNIK J 2013 INFORMATION AND SOFTWARE TECHNOLOGY', 'KOLOZALI S 2013 IEEE TRANSACTIONS ON AUDIO SPEECH AND LANGUAGE PROCESSING', 'MAZANDU G 2013 BMC BIOINFORMATICS', 'BOLLEGALA D 2013 PLOS ONE', 'THORLEUCHTER D 2013 EXPERT SYSTEMS WITH APPLICATIONS', 'DIXON B 2013 ARTIFICIAL INTELLIGENCE IN MEDICINE', 'MARTELL R 2013 CLINICAL THERAPEUTICS', 'FARNAGHI M 2013 COMPUTERS ENVIRONMENT AND URBAN SYSTEMS', 'KHATTAK A 2013 JOURNAL OF INFORMATION SCIENCE AND ENGINEERING', 'ZHANG J 2013 CANADIAN JOURNAL OF INFORMATION AND LIBRARY SCIENCE-REVUE CANADIENNE DES SCIENCES DE L INFORMATION ET DE BIBLIOTHECONOMIE', 'AZZAOUI K 2013 DRUG DISCOVERY TODAY', 'ZHANG C 2013 JOURNAL OF COMPUTER SCIENCE AND TECHNOLOGY', 'ZHANG J 2013 JOURNAL OF COMPUTER SCIENCE AND TECHNOLOGY', 'BOBILLO F 2013 APPLIED SOFT COMPUTING', 'KHAN W 2013 COMPUTING', 'YUE P 2013 ISPRS JOURNAL OF PHOTOGRAMMETRY AND REMOTE SENSING']
                
        self.assertSetEqual(set(ayjid_list),set(expected_ayjid_list), \
                            "AYJID list is not as expected")
        
        # Check for OS Error if there is no such file or directory
        with self.assertRaises(OSError):
            a = rd.wos.from_dir(self.wrong_path)
    
    

    def tearDown(self):
        pass

if __name__ == '__main__':
    unittest.main()



"""
Methods for building networks from bibliographic data.

Each network relies on certain meta data in the :class:`.Paper` associated with
each document. Often we wish to construct a network with nodes representing
these documents and edges representing relationships between those documents,
but this is not always the case.

Where it is the case, it is recommended but not required that nodes are
represented by an identifier from {ayjid, wosid, pmid, doi}. Each has certain
benefits. If the documents to be networked come from a single database source
such as the Web of Science, wosid is most appropriate. If not, using doi
will result in a more accurate, but also more sparse network; while ayjid
will result in a less accurate, but more complete network.

Any type of meta data from the :class:`.Paper` may be used as an identifier,
however.

We use "head" and "tail" nomenclature to refer to the members of a directed
edge (x,y), x -> y, xy, etc. by calling x the "tail" and y the "head".

Modules
```````

.. autosummary::

   authors
   helpers
   papers
   terms
   topics

"""


#import papers
#import authors
#import helpers
#import terms

#Added because of pylint error.
import tethne.networks.papers
import tethne.networks.authors
import tethne.networks.helpers
import tethne.networks.terms
import tethne.networks.topics



#import unittest
#import tethne.crossref as cr
#
#class TestCrossrefQuery(unittest.TestCase):
#    """
#    Tests a sample query with a known DOI number
#    The query is based on an early sample record from Pubmed
#    """
#    def setUp(self):
#        author = 'Griffiths'
#        title = 'Seasonal changes in the carotenoids of the sea urchin'
#        journal = ('Comparative Biochemistry and Physiology Part B: ' +
#                   'Comparative Biochemistry')
#        year = 1976
#        volume = 55
#        fpage = 435
#        self.doi = cr.query(aulast=author, atitle = title, jtitle=journal, 
#                            volume=volume, spage=fpage, date=year)
#    def test_query(self):
#        self.assertEqual(self.doi, '10.1016/0305-0491(76)90317-5')
#
#if __name__ == '__main__':
#    unittest.main()



"""
Methods for generating networks in which papers are vertices.

Methods
```````

.. autosummary::

   author_coupling
   bibliographic_coupling
   cocitation
   direct_citation
   topic_coupling

"""

import networkx as nx
import tethne.utilities as util
import helpers
import operator
import tethne.data as ds

def direct_citation(papers, node_id='ayjid', node_attribs=['date'], **kwargs):
    """
    Create a traditional directed citation network.

    Direct-citation graphs are `directed acyclic graphs`__ in which vertices are
    papers, and each (directed) edge represents a citation of the target
    paper by the source paper. The :func:`.networks.papers.direct_citation`
    method generates both a global citation graph, which includes all cited and
    citing papers, and an internal citation graph that describes only citations
    among papers in the original dataset.

    .. _dag: http://en.wikipedia.org/wiki/Directed_acyclic_graph

    __ dag_

    To generate direct-citation graphs, use the
    :func:`.networks.papers.direct_citation` method. Note the size difference
    between the global and internal citation graphs.

    .. code-block:: python

       >>> gDC, iDC = nt.papers.direct_citation(papers)
       >>> len(gDC)
       5998
       >>> len(iDC)
       163

    ==============     =========================================================
    Element            Description
    ==============     =========================================================
    Node               Papers, represented by node_id.
    Edge               From a paper to a cited reference.
    Edge Attribute     Publication date of the citing paper.
    ==============     =========================================================

    Parameters
    ----------
    papers : list
        A list of :class:`.Paper` instances.

    node_id : int
        A key from :class:`.Paper` to identify the nodes. Default is 'ayjid'.

    node_attribs : list
        List of user provided optional arguments apart from the provided
        positional arguments.

    Returns
    -------
    citation_network : networkx.DiGraph
        Global citation network (all citations).
    citation_network_internal : networkx.DiGraph
        Internal citation network where only the papers in the list are nodes in
        the network.

    Raises
    ------
    KeyError : If node_id is not present in the meta_list.
    """
    citation_network = nx.DiGraph(type='citations')
    citation_network_internal = nx.DiGraph(type='citations')

    # Check node_id validity.
    meta_dict = ds.Paper()
    meta_keys = meta_dict.keys()
    if node_id not in meta_keys:
        raise KeyError('node_id:' + node_id + 'is not in the set of' +
                       'meta_keys')

    for entry in papers:
        # Check the head.
        head_has_id = True
        if entry[node_id] is None:
            head_has_id = False

        if head_has_id:
            # Then create node to both global and internal networks.
            node_attrib_dict = util.subdict(entry, node_attribs)
            citation_network.add_node(entry[node_id], node_attrib_dict)
            citation_network_internal.add_node(entry[node_id],
                                               node_attrib_dict)
        if entry['citations'] is not None:
            for citation in entry['citations']:
                # Check the tail.
                tail_has_id = True
                if citation[node_id] is None:
                    tail_has_id = False

                if tail_has_id:
                    # Then create node to global but not internal network.
                    node_attrib_dict = util.subdict(citation, node_attribs)
                    citation_network.add_node(citation[node_id],
                                              node_attrib_dict)

                if head_has_id and tail_has_id:
                    # Then draw an edge in the network.
                    citation_network.add_edge(entry[node_id],
                                              citation[node_id],
                                              date=entry['date'])

                    # And check if it can be added to the internal network, too.
                    if (util.contains (papers,
                                       lambda wos_obj:
                                       wos_obj[node_id] == citation[node_id])):
                        citation_network_internal.add_edge(
                            entry[node_id],
                            citation[node_id],
                            date=entry['date'])

    # Checking if both the graphs are Directed Acyclic Graphs.
    if not nx.is_directed_acyclic_graph(citation_network):
        raise nx.NetworkXError("Citation graph is not a DAG.")
    elif not nx.is_directed_acyclic_graph(citation_network_internal):
        raise nx.NetworkXError("Internal citation graph is not a DAG.")
    else:
        return citation_network, citation_network_internal

def bibliographic_coupling(papers, citation_id='ayjid', threshold=1,
                           node_id='ayjid', node_attribs=['date'],
                           weighted=False, **kwargs):
    """
    Generate a bibliographic coupling network.

    Two papers are **bibliographically coupled** when they both cite the same,
    third, paper. You can generate a bibliographic coupling network using the
    :func:`.networks.papers.bibliographic_coupling` method.

    .. code-block:: python

       >>> BC = nt.papers.bibliographic_coupling(papers)
       >>> BC
       <networkx.classes.graph.Graph object at 0x102eec710>

    Especially when working with large datasets, or disciplinarily narrow
    literatures, it is usually helpful to set a minimum number of shared
    citations required for two papers to be coupled. You can do this by setting
    the **`threshold`** parameter.

    .. code-block:: python

       >>> BC = nt.papers.bibliographic_coupling(papers, threshold=1)
       >>> len(BC.edges())
       1216
       >>> BC = nt.papers.bibliographic_coupling(papers, threshold=2)
       >>> len(BC.edges())
       542

    ===============    =========================================================
    Element            Description
    ===============    =========================================================
    Node               Papers represented by node_id.
    Node Attributes    node_attribs in :class:`.Paper`
    Edge               (a,b) in E(G) if a and b share x citations where x >=
                       threshold.
    Edge Attributes    overlap: the number of citations shared
    ===============    =========================================================


    Parameters
    ----------
    papers : list
        A list of wos_objects.
    citation_id: string
        A key from :class:`.Paper` to identify the citation overlaps.  Default
        is 'ayjid'.
    threshold : int
        Minimum number of shared citations to consider two papers "coupled".
    node_id : string
        Field in :class:`.Paper` used to identify the nodes. Default is 'ayjid'.
    node_attribs : list
        List of fields in :class:`.Paper` to include as node attributes in
        graph.
    weighted : bool
        If True, edge attribute `overlap` is a float in {0-1} calculated as
        :math:`\cfrac{N_{ij}}{\sqrt{N_{i}N_{j}}}` where :math:`N_{i}` and
        :math:`N_{j}` are the number of references in :class:`.Paper` *i* and
        *j*, respectively, and :math:`N_{ij}` is the number of references
        shared by papers *i* and *j*.

    Returns
    -------
    bcoupling : networkx.Graph
        A bibliographic coupling network.

    Raises
    ------
    KeyError : Raised when citation_id is not present in the meta_list.

    Notes
    -----
    Lists cannot be attributes? causing errors for both gexf and graphml also
    nodes cannot be none.
    """

    bcoupling = nx.Graph(type='biblio_coupling')

    # Validate identifiers.
    meta_dict = ds.Paper()
    meta_keys = meta_dict.keys()
    if node_id not in meta_keys:
        raise KeyError('node_id' + node_id + ' is not a meta_dict key.')

    # 'citations' is the only invalid meta_key for citation_id
    meta_keys.remove('citations')
    if citation_id not in meta_keys:
        raise KeyError('citation_id' + citation_id + ' is not a meta_dict' +
                       ' key or otherwise cannot be used to detect citation' +
                       ' overlap.')

    for i in xrange(len(papers)):
        # Make a list of citation_id's for each paper...
        i_list = []
        if papers[i]['citations'] is not None:
            for citation in papers[i]['citations']:
                i_list.append(citation[citation_id])

        # ...and construct that paper's node.
        node_i_attribs = util.subdict(papers[i], node_attribs)

        for j in xrange(i+1, len(papers)):
            # Make a list of citation_id's for each paper...
            j_list = []
            if papers[j]['citations'] is not None:
                for citation in papers[j]['citations']:
                    j_list.append(citation[citation_id])

            # ...and construct that paper's node.
            node_j_attribs = util.subdict(papers[j], node_attribs)

            # Add nodes and edge if the citation overlap is sufficiently high.
            overlap = util.overlap(i_list, j_list)

            if weighted:
                if len(overlap) > 0:
                    w = (float(len(i_list)) * float(len(j_list)))**0.5
                    similarity = float(len(overlap)) / w
                else:
                    similarity = 0
            else:
                similarity = len(overlap)

            if similarity >= threshold:
                bcoupling.add_node(papers[i][node_id], node_i_attribs)
                bcoupling.add_node(papers[j][node_id], node_j_attribs)
                #nx.set_node_attributes(bcoupling,"",node_i_attribs)

                bcoupling.add_edge(papers[i][node_id],
                                   papers[j][node_id],
                                   similarity=similarity)
    return bcoupling

def cocitation(papers, threshold=1, node_id='ayjid', topn=None, verbose=False,\
                node_attribs=['date'], **kwargs):
    """
    Generate a cocitation network.

    A **cocitation network** is a network in which vertices are papers, and
    edges indicate that two papers were cited by the same third paper.
    `CiteSpace <http://cluster.cis.drexel.edu/~cchen/citespace/doc/jasist2006.pdf>`_
    is a popular desktop application for co-citation analysis, and you can read
    about the theory behind it
    `here <http://cluster.cis.drexel.edu/~cchen/citespace/>`_. Co-citation
    analysis is generally performed with a temporal component, so building a
    :class:`.GraphCollection` from a :class`.DataCollection` sliced by ``date``
    is recommended.

    You can generate a co-citation network using the
    :func:`.networks.papers.cocitation` method:

    .. code-block:: python

       >>> CC = nt.papers.cocitation(papers)
       >>> CC
       <networkx.classes.graph.Graph object at 0x102eec790>

    For large datasets, you may wish to set a minimum number of co-citations
    required for an edge between two papers Keep in mind that all of the
    references in a single paper are co-cited once, so a threshold of at least
    2 is prudent. Note the dramatic decrease in the number of edges when the
    threshold is changed from 2 to 3.

    .. code-block:: python

       >>> CC = nt.papers.cocitation(papers, threshold=2)
       >>> len(CC.edges())
       8889
       >>> CC = nt.papers.cocitation(papers, threshold=3)
       >>> len(CC.edges())
       1493

    ===============    =========================================================
    Element            Description
    ===============    =========================================================
    Node               Cited papers represented by :class:`.Paper` ayjid.
    Edge               (a, b) if a and b are cited by the same paper.
    Edge Attributes    weight: number of times two papers are co-cited
                       together.
    ===============    =========================================================

    Parameters
    ----------
    papers : list
        a list of :class:`.Paper` objects.
    threshold : int
        Minimum number of co-citations required to create an edge.
    topn : int or float, or None
        If provided, only the topn (int) or topn percent (float) most cited
        papers will be included in the cocitation network. If None (default),
        network will include all cited papers (NOTE: this can cause severe
        memory consumption for even moderately-sized datasets).
    verbose : bool
        If True, prints status messages.

    Returns
    -------
    cocitation : networkx.Graph
        A cocitation network.

    """

    cocitation_graph = nx.Graph(type='cocitation')

    # We'll use tuples as keys. Values are the number of times each pair
    #  of papers is co-cited.
    cocitations = {}
    citations_count = {}

    # 61670334: networks.citations.cocitation should have a "top cited"
    #  parameter.
    if topn is not None:
        parents,include,citations_count = helpers.top_parents(papers, topn=topn)
        N = len(include)
    else:
        citations_count = helpers.citation_count(papers)
        N = len(citations_count.keys())

    if verbose:
        print "Generating a cocitation network with " + str(N) + " nodes..."

    for paper in papers:
        if paper['citations'] is not None:  # Some papers don't have citations.
            n = len(paper['citations'])
            for i in xrange(0, n):
                paper_i = paper['citations'][i]['ayjid']

                if topn is not None and paper_i not in include:
                    pass
                else:
                    for j in xrange(i+1, n):
                        paper_j = paper['citations'][j]['ayjid']

                        if topn is not None and paper_j not in include:
                            pass
                        else:

                            pp = ( paper_i, paper_j )
                            pp_inv = ( paper_j, paper_i )

                            try: # Have these papers been co-cited before?
                                cocitations[pp] += 1
                            except KeyError:
                                try: # Maybe in opposite order?
                                    cocitations[pp_inv] += 1
                                except KeyError:
                                    # First time these papers are co-cited.
                                    cocitations[pp] = 1

    if verbose:
        print "Co-citation matrix generated, building Graph..."

    for key , val in cocitations.iteritems():
        if val >= threshold: # and key[0] in include and key[1] in include:
            cocitation_graph.add_edge(key[0], key[1], weight=val)

    if verbose:
        print "Done building co-citation graph, adding attributes..."

    # 62657522: Nodes in co-citation graph should have attribute containing
    #  number of citations.
    n_cit = { k:v for k,v in citations_count.iteritems()
                if k in cocitation_graph.nodes() }
    nx.set_node_attributes( cocitation_graph, 'citations', n_cit )

    return cocitation_graph

def author_coupling(papers, threshold=1, node_attribs=['date'],
                                                     node_id='ayjid', **kwargs):
    """
    Vertices are papers and edges indicates shared authorship.

    ===============    =========================================================
    Element            Description
    ===============    =========================================================
    Node               Papers, represented by node_id.
    Edge               (a,b) in E(G) if a and b share x authors and x >=
                       threshold
    Edge Attributes    overlap: the value of x (above).
    ===============    =========================================================

    Parameters
    ----------
    papers : list
        A list of :class:`.Paper`
    threshold : int
        Minimum number of co-citations required to draw an edge between two
        authors.
    node_id : string
        Field in :class:`.Paper` used to identify nodes.
    node_attribs : list
        List of fields in :class:`.Paper` to include as node attributes in
        graph.

    Returns
    -------
    acoupling : networkx.Graph
        An author-coupling network.

    """
    acoupling = nx.Graph(type='author_coupling')

    for i in xrange(len(papers)):
        #define last name first initial name lists for each paper
        name_list_i = util.concat_list(papers[i]['aulast'],
                                       papers[i]['auinit'],
                                       ' ')

        #create nodes
        node_attrib_dict = util.subdict(papers[i], node_attribs)

        acoupling.add_node(papers[i][node_id], node_attrib_dict)

        for j in xrange(i+1, len(papers)):
            #define last name first initial name lists for each paper
            name_list_j = util.concat_list(papers[j]['aulast'],
                                           papers[j]['auinit'],
                                           ' ')

            #create nodes
            node_attrib_dict = util.subdict(papers[j], node_attribs)
            acoupling.add_node(papers[j][node_id], node_attrib_dict)

            #draw edges as appropriate
            overlap = util.overlap(name_list_i, name_list_j)

            if len(overlap) >= threshold:
                acoupling.add_edge(papers[i][node_id],
                                   papers[j][node_id],
                                   overlap=len(overlap))
    return acoupling

def topic_coupling(papers, threshold=0.7, node_id='ayjid', **kwargs):
    """
    Two papers are coupled if they both contain a shared topic above threshold.

    ===============    =========================================================
    Element            Description
    ===============    =========================================================
    Node               Papers, represented by node_id.
    Edge               (a,b) in E(G) if a and b share >= 1 topics with
                       proportion >= threshold in both a and b.
    Edge Attributes    weight: combined mean proportion of each shared topic.
                       topics: list of shared topics.
    ===============    =========================================================

    Parameters
    ----------
    papers : list
        A list of :class:`.Paper`
    threshold : float
        Minimum representation of a topic in each paper.
    node_id : string
        Field in :class:`.Paper` used to identify nodes.

    Returns
    -------
    tc : networkx.Graph
        A topic-coupling network.
    """
    for i in xrange(len(papers)):
        t_i = papers[i]['topics'][0]    # Topic vector for i.
        for j in xrange(i+1,len(papers)):
            t_j = papers[i]['topics'][0]    # Topic vector for j.

            Z = t_i.shape[0]
            for z in xrange(Z):
                if t_i[z] >= threshold and t_j[z] >= threshold:
                    try:  # Add topic and mean of representation in i and j.
                        edges[(i,j)].append( (z,(t_i[z]+t_j[z])/2) )
                    except KeyError:
                        edges[(i,j)] = [ (z,(t_i[z]+t_j[z])/2) ]

    tc = nx.Graph()

    # Combine means of representations into a single edge weight in {0-1}.
    for e, topics in edges.iteritems():
        weight = sum([ t[1] for t in topics ] ) / t_i.shape[0]
        i_id = papers[e[0]][node_id]
        j_id = papers[e[1]][node_id]
        tc.add_edge(i_id, j_id, weight=weight, topics=[t[0] for t in topics])

    return tc



"""
Methods for generating networks in which authors are vertices.

Methods
```````

.. autosummary::

   author_cocitation
   author_coinstitution
   author_institution
   author_papers
   coauthors

"""

import networkx as nx
import tethne.utilities as util
import tethne.data as ds
from collections import defaultdict

# MACRO for printing the 'print' statement values.
# 0 prints nothing in the console.
# 1 prints all print statements in the console.
DEBUG = 0

def author_papers(papers, node_id='ayjid', paper_attribs=[], **kwargs):
    """
    Generate an author_papers network NetworkX directed graph.

    ==============     =========================================================
    Element            Description
    ==============     =========================================================
    Node               Two kinds of nodes with distinguishing "type" attributes:
                       * type = paper    - a paper in papers
                       * type = person   - a person in papers
                       Papers node attributes defined by paper_attribs.
    Edge               Directed, Author -> his/her Paper.
    ==============     =========================================================

    Parameters
    ----------
    papers : list
        A list of wos_objects.
    node_id : string
        A key from :class:`.Paper` used to identify the nodes.
    paper_attribs : list
        List of user-provided optional arguments apart from the provided
        positional arguments.

    Returns
    -------
    author_papers_graph : networkx.DiGraph
        A DiGraph 'author_papers_graph'.

    Raises
    ------
    KeyError : Raised when node_id is not present in Papers.

    """
    author_papers_graph = nx.DiGraph(type='author_papers')

    # Validate node_id.
    meta_dict = ds.Paper()
    meta_keys = meta_dict.keys()
    meta_keys.remove('citations')
    if node_id not in meta_keys:
        raise KeyError('node_id' + node_id + ' cannot be used to identify' +
                       ' papers.')
    for entry in papers:
        # Define paper_attribute dictionary.
        paper_attrib_dict = util.subdict(entry, paper_attribs)
        paper_attrib_dict['type'] = 'paper'
        # Add paper node with attributes.
        author_papers_graph.add_node(entry[node_id], paper_attrib_dict)

        authors = util.concat_list(entry['aulast'], entry['auinit'], ' ')
        for i in xrange(len(authors)):
            # Add person node.
            author_papers_graph.add_node(authors[i], type="person")
            # Draw edges.
            author_papers_graph.add_edge(authors[i], entry[node_id],
                                   date=entry['date'])

    return author_papers_graph

def coauthors(papers, threshold=1, edge_attribs=['ayjid'], **kwargs):
    """
    Generate a co-author network.

    As the name suggests, edges are drawn between two author-vertices in the
    case that those authors published a paper together. Co-authorship networks
    are popular models for studying patterns of collaboration in scientific
    communities.

    To generate a co-authorship network, use the
    :func:`.networks.authors.coauthors` method:

    .. code-block:: python

       >>> CA = nt.authors.coauthors(papers)
       >>> CA
       <networkx.classes.multigraph.MultiGraph object at 0x101705650>

    ==============     =========================================================
    Element            Description
    ==============     =========================================================
    Node               Author name.
    Edges              (a,b) in E(G) if a and b are coauthors on the same paper.
    ==============     =========================================================

    Parameters
    ----------
    papers : list
        A list of :class:`Paper` instances.
    threshold : int
        Minimum number of co-citations required for an edge. (default: 1)
    edge_attribs : list
        List of edge_attributes specifying which :class:`.Paper` keys (from the
        co-authored paper) to use as edge attributes. (default: ['ayjid'])

    Returns
    -------
    coauthors_graph : networkx.MultiGraph
        A co-authorship network.

    """

    # TODO: Check whether papers contains :class:`.Paper` instances, and raise
    #  an exception if not.

    coauthors_graph = nx.Graph(type='coauthors')
    edge_att = {}
    #edge_listdict={}
    coauthor_dict = {}
    for entry in papers:
        if entry['aulast'] is not None:
            # edge_att dictionary has the atributes given by user input
            #  for any edges that get added
            edge_att = util.subdict(entry, edge_attribs)
            # make a new list of aulast, auinit names
            full_names = util.concat_list(entry['aulast'],
                                          entry['auinit'],
                                          ' ')
            for a in xrange(len(full_names)):
                for b in xrange(a+1 , len(entry['aulast'])):
                    # (author_a,author_b) tuple is key for coauthor_dict.
                    authors = full_names[a], full_names[b]
                    authors_inv = full_names[b], full_names[a]

                    if authors in coauthor_dict.keys():
                        for key,value in edge_att.iteritems():
                            coauthor_dict[authors][key].append(value)
                        coauthor_dict[authors]['weight'] += 1

                    # Author names may be in inverse order.
                    elif authors_inv in coauthor_dict.keys():
                        for key,value in edge_att.iteritems():
                            coauthor_dict[authors_inv][key].append(value)
                        coauthor_dict[authors_inv]['weight'] += 1

                    # First encounter for this author-pair.
                    else:
                        coauthor_dict[authors] = { k:[v]
                                                    for k,v
                                                    in edge_att.iteritems() }
                        coauthor_dict[authors]['weight'] = 1

    # Add edges with specified edge attributes.
    for key, val in coauthor_dict.iteritems():
        if val['weight'] >= threshold:
            coauthors_graph.add_edge(key[0], key[1], attr_dict=val)

    return coauthors_graph

def author_institution(Papers, edge_attribs=[], **kwargs):
    """
    Generate a bi-partite graph with people and institutions as vertices, and
    edges indicating affiliation. This may be slightly ambiguous for WoS data
    where num authors != num institutions.

    ==============     =========================================================
    Element            Description
    ==============     =========================================================
    Node               Author name.
    Edge               (a,b) in E(G) if a and b are authors on the same paper.
    ==============     =========================================================

    Parameters
    ----------
    Papers : list
        A list of :class:`.Paper` instances.
    edge_attribs : list
        List of edge_attributes specifying which :class:`.Paper` keys (from the
        authored paper) to use as edge attributes. For example, the 'date' key
        in :class:`.Paper` .

    Returns
    -------
    author_institution_graph : networkx.MultiGraph
        A graph describing institutional affiliations of authors in the corpus.
    """

    author_institution_graph = nx.MultiGraph(type='author_institution')
    #The Field in Papers which corresponds to authors and affliated institutions
    # is "institutions"
    # { 'institutions' : { Authors:[institutions_list]}}
    for paper in Papers:
        if paper['institutions'] is not None:
            auth_inst = paper['institutions']
            edge_attrib_dict = util.subdict(paper, edge_attribs)
            authors = auth_inst.keys()
            for au in authors:
                #add node of type 'author'
                author_institution_graph.add_node(au, type='author')
                ins_list = auth_inst[au]
                for ins_str in ins_list:
                  #add node of type 'institutions'
                    author_institution_graph.add_node(ins_str, \
                                                      type='institution')
                  #print au ,'---->' , ins_str
                    author_institution_graph.add_edge(au, ins_str, \
                                              attr_dict=edge_attrib_dict)


    return author_institution_graph

def author_coinstitution(Papers, threshold=1, **kwargs):
    """
    Generate a co-institution graph, where edges indicate shared affiliation.

    Some bibliographic datasets, including data from the Web of Science,
    includes the institutional affiliations of authors. In a co-institution
    graph, two authors (vertices) have an edge between them if they share an
    institutional affiliation in the dataset. Note that data about institutional
    affiliations varies in the WoS database so this will yield more reliable
    results for more recent publications.

    To generate a co-institution network, use the
    :func:`.networks.authors.author_coinstitution` method:

    .. code-block:: python

       >>> ACI = nt.authors.author_coinstitution(papers)
       >>> ACI
       <networkx.classes.graph.Graph object at 0x106571190>

    ==============     =========================================================
    Element            Description
    ==============     =========================================================
    Node               Authors.
    Node Attribute     type (string). 'author' or 'institution'.
    Edges              (a, b) where a and b are affiliated with the same
                       institution.
    Edge attribute     overlap (int). number of shared institutions.
    ==============     =========================================================

    Parameters
    ----------
    Papers : list
        A list of wos_objects.
    threshold : int
        Minimum institutional overlap required for an edge.

    Returns
    -------
    coinstitution : NetworkX :class:`.graph`
        A coinstitution network.

    """
    coinstitution = nx.Graph(type='author_coinstitution')


    # The Field in Papers which corresponds to the affiliation is "institutions"
    #  { 'institutions' : { Authors:[institutions_list]}}
    author_institutions = {}  # keys: author names, values: list of institutions
    for paper in Papers:
        if paper['institutions'] is not None:
            for key, value in paper['institutions'].iteritems():
                try:
                    author_institutions[key] += value
                except KeyError:
                    author_institutions[key] = value
        authors = author_institutions.keys()
        for i in xrange(len(authors)):
            #coinstitution.add_node(authors[i],type ='author')
            for j in xrange(len(authors)):
                if i != j:
                     #compare 2 author dict elements
                    overlap = (set(author_institutions[authors[i]])
                                &
                                set(author_institutions[authors[j]]))
                    if len(overlap) >= threshold:
                        #commented these because 'add_edge' adds nodes aswell.
                        #coinstitution.add_node(authors[i],type ='author')
                        #coinstitution.add_node(authors[j],type ='author')
                        #print authors[i] + "->" + authors[j]
                        coinstitution.add_edge(authors[i], authors[j], \
                                               overlap=len(overlap))
                    else :
                        pass
        #62809656
        attribs_dict = {}
        for node in coinstitution.nodes():
            attribs_dict[node] = 'author'
        nx.set_node_attributes( coinstitution, 'type', attribs_dict )


    return coinstitution

def author_cocitation(papers, threshold=1, **kwargs):
    """
    Generates an author co-citation network; edges indicate co-citation of
    authors' papers.

    Similar to :func:`.papers.cocitation`\, except that vertices are authors
    rather than papers. To generate an author co-citation network, use the
    :func:`.networks.authors.author_cocitation` method:

    .. code-block:: python

       >>> ACC = nt.authors.author_cocitation(papers)
       >>> ACC
       <networkx.classes.graph.Graph object at 0x106571190>

    ==============     =========================================================
    Element            Description
    ==============     =========================================================
    Nodes              Author name.
    Edge               (a, b) if a and b are referenced by the same paper in
                       papers
    Edge attribute     'weight', the number of papers that co-cite a and b.
    ==============     =========================================================

    Parameters
    ----------
    papers : list
        a list of :class:`.Paper` objects.
    threshold : int
        Minimum number of co-citations required to create an edge between
        authors.

    Returns
    -------
    cocitation : :class:`.networkx.Graph`
        A cocitation network.

    """

    author_cocitations = nx.Graph(type='author_cocitation')

    # We'll use tuples as keys. Values are the number of times each pair
    # of 2 authors is co-cited.

    cocitations = {}
    delim = ' '

    for paper in papers:
        # Some papers don't have citations.
        if paper['citations'] is not None:
            # n is the number of papers in the provided list of Papers.
            n = len(paper['citations'])
            found_authors = []  # To avoid extra incrementation of author pairs.
            if n > 1:   # No point in proceeding if there is only one citation.
                for i in xrange(0, n):

                    # al_i_str is the author i's last name.
                    # converting list to str
                    al_i_str = ''.join(map(str, \
                                            (paper['citations'][i]['aulast'])))

                    # ai_i_str is the author i's first name
                    # converting list to str

                    ai_i_str = \
                        ''.join(map(str,(paper['citations'][i]['auinit'])))

                    # Making it a tuple,that it becomes key for cocitations dict
                    author_i_str = al_i_str + delim + ai_i_str

                    # Start inner loop at i+1,\
                    # to avoid redundancy and self-loops.

                    for j in xrange(i+1, n):
                        # al_j_str is the last name of author j
                        al_j_str = ''.join(map(str, \
                                            (paper['citations'][j]['aulast'])))

                        # ai_j_str is the author j's first name
                        # converting list to str

                        ai_j_str = ''.join(map(str, \
                                           (paper['citations'][j]['auinit'])))

                        # Making it a tuple so that it becomes the key for
                        # cocitations dict
                        author_j_str = al_j_str + delim + ai_j_str

                        # 2 tuples which are going to be the keys of the dict.
                        authors_pair = (author_i_str.upper(), \
                                                author_j_str.upper())
                        authors_pair_inv = (author_j_str.upper(), \
                                                author_i_str.upper())

                        if DEBUG:
                            print "Authors Pairs: ", authors_pair
                            print "Authors Pairs inv: ", authors_pair_inv

                        # Have these authors been co-cited before?
                        try:
                            # check if author pair is not already \
                            # in the list and
                            # if the pair and inverse are not same. This is done
                            # to avoid drawing edges between same authors(nodes)

                            if (authors_pair not in found_authors
                                    and (authors_pair != authors_pair_inv)):
                                cocitations[authors_pair] += 1
                                found_authors.append(authors_pair)

                        except KeyError:
                            try: # May have been entered in opposite order.
                                if (authors_pair_inv not in found_authors
                                    and (authors_pair != authors_pair_inv)):
                                    cocitations[authors_pair_inv] += 1
                                    found_authors.append(authors_pair_inv)
                                # Networkx will ignore add_node
                                # if those nodes are already present
                            except KeyError:
                                # First time these papers have been co-cited.
                                cocitations[authors_pair] = 1
                                found_authors.append(authors_pair)

    # Create the network.
    for key, val in cocitations.iteritems():
        # If the weight is greater or equal to the user I/P threshold
        if val >= threshold :
            # Add edge between the 2 co-cited authors
            author_cocitations.add_edge(key[0], key[1], weight=val)

    return author_cocitations



from distutils.core import setup

DISTNAME = 'tethne'
MAINTAINER = 'Erick Peirson'
MAINTAINER_EMAIL = 'erick [dot] peirson [at] asu [dot] edu'
DESCRIPTION = ('Python library for analyzing bibliographic data')
LICENSE = 'GNU GPL 3' 
URL = 'https://github.com/erickpeirson/tethne' 
VERSION = '0.3.0-alpha'
PACKAGES = [ 'tethne','tethne.analyze','tethne.networks','tethne.readers', \
             'tethne.utilities', 'tethne.writers', 'tethne.matrices',
             'testsuite' ]

setup(name=DISTNAME,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      description=DESCRIPTION,
      license=LICENSE,
      url=URL,
      version=VERSION,
      packages = PACKAGES,
      )


"""
Helper functions for :mod:`tethne.networks` .
"""
def contains(l, f):
    """
    Searches list l for a pattern specified in a lambda function f.
    """
    for x in l:
        if f(x):
            return True
    return False


def overlap(listA, listB):
    """
    Return list of objects shared by listA, listB.
    """
    if (listA is None) or (listB is None):
        return []
    else:
        return list(set(listA) & set(listB))


def subdict(super_dict, keys):
    """
    Returns a subset of the super_dict with the specified keys.
    """
    sub_dict = {}
    valid_keys = super_dict.keys()
    for key in keys:
        if key in valid_keys:
            sub_dict[key] = super_dict[key]

    return sub_dict


def attribs_to_string(attrib_dict, keys):
    """
    A more specific version of the subdict utility aimed at handling
    node and edge attribute dictionaries for NetworkX file formats such as
    gexf (which does not allow attributes to have a list type) by making
    them writable in those formats
    """
    for key, value in attrib_dict.iteritems():
        if (isinstance(value, list) or isinstance(value, dict) or
            isinstance(value, tuple)):
            attrib_dict[key] = str(value)

    return attrib_dict


def concat_list(listA, listB, delim=' '):
    """
    Concatenate list elements pair-wise with the delim character
    Returns the concatenated list
    Raises index error if lists are not parallel
    """

    # Lists must be of equal length.
    if len(listA) != len(listB):
        raise IndexError('Input lists are not parallel.')

    # Concatenate lists.
    listC = []
    for i in xrange(len(listA)):
        app = listA[i] + delim + listB[i]
        listC.append(app)

    return listC

def strip_non_ascii(string):
    """
    Returns the string without non-ASCII characters.

    Parameters
    ----------
    string : string
        A string that may contain non-ASCII characters.

    Returns
    -------
    clean_string : string
        A string that does not contain non-ASCII characters.

    """
    stripped = (c for c in string if 0 < ord(c) < 127)
    clean_string = ''.join(stripped)
    return clean_string

def dict_from_node(node, recursive=False):
    """
    Converts ElementTree node to a dictionary.

    Parameters
    ----------
    node : ElementTree node
    recursive : boolean
        If recursive=False, the value of any field with children will be the
        number of children.

    Returns
    -------
    dict : nested dictionary.
        Tags as keys and values as values. Sub-elements that occur multiple
        times in an element are contained in a list.
    """

    dict = {}
    for snode in node:
        if len(snode) > 0:
            if recursive:
                # Will drill down until len(snode) <= 0.
                value = dict_from_node(snode, True)
            else:
                value = len(snode)
        elif snode.text is not None:
            value = strip_non_ascii(snode.text)
        else:
            value = ''

        if snode.tag in dict.keys():    # If there are multiple subelements
                                        #  with the same tag, then the value
                                        #  of the element should be a list
                                        #  rather than a dict.
            if type(dict[snode.tag]) is list:   # If a list has already been
                                                #  started, just append to
                                                #  it.
                dict[snode.tag].append(value)
            else:
                dict[snode.tag] = [ dict[snode.tag], value ]
        else:
            dict[snode.tag] = value     # Default behavior.
    return dict
    
class Dictionary:
    """
    A two-way index for integer/string pairs.
    """
    def __init__(self):
        self.by_str = {}
        self.by_int = {}
        
    def __setitem__(self, key, value):
        if type(key) == str:
            self.by_str[key] = value
            self.by_int[value] = key
        if type(key) == int:
            self.by_int[key] = value
            self.by_str[value] = key
    
    def __getitem__(self, key):
        if type(key) == str:
            return self.by_str[key]
        if type(key) == int:
            return self.by_int[key]


"""
Write :class:`.GraphCollection` to a structured data format.

.. autosummary::

   to_dxgmml
   
"""

import networkx as nx
import pickle as pk

def to_dxgmml(C, path): # [#61510094]
    """
    Writes a :class:`.GraphCollection` to 
    `dynamic XGMML. <https://code.google.com/p/dynnetwork/wiki/DynamicXGMML>`_.
    
    Dynamic XGMML is a schema for describing dynamic networks in Cytoscape 3.0.
    This method assumes that `Graph` indices are orderable points in time 
    (e.g. years). The "start" and "end" of each node and edge are determined by 
    periods of consecutive appearance in the :class:`.GraphCollection` . Node 
    and edge attributes are defined for each `Graph`. in the 
    :class:`.GraphCollection`.
    
    For example, to build and visualize an evolving co-citation network:
    
    .. code-block:: python

       >>> # Load some data.
       >>> import tethne.readers as rd
       >>> papers = rd.wos.read(datapath)

       >>> # Build a DataCollection, and slice it temporally using a
       >>> #  4-year sliding time-window.
       >>> from tethne.data import DataCollection, GraphCollection
       >>> D = DataCollection(papers)
       >>> D.slice('date', 'time_window', window_size=4)

       >>> # Generate a GraphCollection of co-citation graphs.
       >>> from tethne.builders import paperCollectionBuilder
       >>> builder = paperCollectionBuilder(D)
       >>> C = builder.build('date', 'cocitation', threshold=2)

       >>> # Write the GraphCollection as a dynamic network.
       >>> import tethne.writers as wr
       >>> wr.collection.to_dxgmml(C, "/path/to/network.xgmml")
    
    Parameters
    ----------
    C : :class:`.GraphCollection`
        The :class:`.GraphCollection` to be written to XGMML.
    path : str
        Path to file to be written. Will be created/overwritten.

    Notes
    -----
    Period start and end dates in this method are inclusive, whereas XGMML end
    dates are exclusive. Hence +1 is added to all end dates when writing XGMML.
    """

    nodes = {}
    for n in C.nodes():
        nodes[n] = { 'periods' : [] }   # Each period will be a dict with
                                        #  'start' and 'end' values.
    edges = {}
    for e in C.edges():
        edges[e] = { 'periods' : [] }

    # Build node list.
    current = []
    for k in sorted(C.graphs.keys()):
        G = _strip_list_attributes(C[k])
        preceding = current
        current = []
        for n in G.nodes(data=True):
            if n[0] not in preceding:   # Looking for gaps in presence of node.
                nodes[n[0]]['periods'].append( { 'start': k, 'end': k } )
            else:
                if k > nodes[n[0]]['periods'][-1]['end']:
                    nodes[n[0]]['periods'][-1]['end'] = k
            current.append(n[0])

            nodes[n[0]][k] = {}
            for attr, value in n[1].iteritems():
                nodes[n[0]][k][attr] = value

    # Build edge list.
    current = []
    for k in sorted(C.graphs.keys()):
        G = _strip_list_attributes(C[k])
        preceding = current
        current = []
        for e in G.edges(data=True):
            e_key = (e[0], e[1])
            if e_key not in preceding:   # Looking for gaps in presence of edge.
                edges[e_key]['periods'].append( { 'start': k, 'end': k } )
            else:
                if k > edges[e_key]['periods'][-1]['end']:
                    edges[e_key]['periods'][-1]['end'] = k
            current.append(e_key)

            edges[e_key][k] = {}
            for attr, value in e[2].iteritems():
                if type(value) is str:
                    value = value.replace("&", "&amp;").replace('"', '')
                edges[e_key][k][attr] = value

    # Write graph to XGMML.
    nst = '\t<node label="{0}" id="{0}" start="{1}" end="{2}">\n'
    ast = '\t\t<att name="{0}" type="{1}" value="{2}" start="{3}" end="{4}"/>\n'
    
    with open(path, "w") as f:
        f.write('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n')
        f.write('<graph>\n')
        for n in nodes.keys():
            for period in nodes[n]['periods']:
                label = str(n).replace("&", "&amp;").replace('"', '')
                
                f.write(nst.format(label, period['start'], period['end']+1))

                for i in sorted(nodes[n].keys()):
                    if period['start'] <= i <= period['end']:
                        for attr, value in nodes[n][i].iteritems():
                            # Type names are slightly different in XGMML.
                            if type(value) is str: dtype = 'string'
                            if type(value) is int: dtype = 'integer'
                            if type(value) is float: dtype = 'real'
                            attr = str(attr).replace("&", "&amp;")
                            f.write(ast.format(attr, dtype, value, i, i+1))
                f.write('\t</node>\n')

        for e in edges.keys():
            for period in edges[e]['periods']:
                src = str(e[0]).replace("&", "&amp;").replace('"', '')
                tgt = str(e[1]).replace("&", "&amp;").replace('"', '')
                start = str(period['start'])
                end = str(period['end']+1)
                f.write('\t<edge source="' + src + '" target="' + tgt \
                            + '" start="'+ start + '" end="' + end + '">\n')

                for i in sorted(edges[e].keys()):
                    if period['start'] <= i <= period['end']:
                        for attr, value in edges[e][i].iteritems():
                            # Type names are slightly different in XGMML.
                            if type(value) is str: dtype = 'string'
                            if type(value) is int: dtype = 'integer'
                            if type(value) is float: dtype = 'real'
                            f.write('\t\t<att name="'+str(attr)+'" type="'\
                                    +dtype+'" value="'+str(value)+'" start="'\
                                    +str(i)+'" end="'\
                                    +str(i+1)+'" />\n'.replace("&", "&amp;"))
                f.write('\t</edge>\n')
        f.write('</graph>')
        
def _strip_list_attributes(G):
    for n in G.nodes(data=True):
        for k,v in n[1].iteritems():
            if type(v) is list:
                G.node[n[0]][k] = str(v)
    for e in G.edges(data=True):
        for k,v in e[2].iteritems():
            if type(v) is list:
                G.edge[e[0]][e[1]][k] = str(v)

    return G        



import unittest
import tethne.data as ds
import networkx as nx
import numpy as np
import pickle as pk
import tethne.networks as nt
import tethne.readers as rd

class TestPaper(unittest.TestCase):
    """
    Tests for tethne.data.Paper class.
    
    Notes
    -----
    TODO: Implement this.
    """
    
    def setUp(self):
        pass
    
    def test_setitem(self):
        """
        Test of tethne.data.Paper.__setitem__ .
        """    
        pass
    
    def test_getitem(self):
        """
        Test of tethne.data.Paper.__getitem__ .
        """
        pass
    
    def tearDown(self):
        pass

class TestGraphCollection(unittest.TestCase):
    """
    Tests methods for tethne.data.GraphCollection class.
    """
    
    
    def setUp(self):
        self.collection = ds.GraphCollection()
        self.correct_filename = "../testsuite/testout/graphcollections1.txt"
        self.wrong_filename = "../testsuite/asdad123e2e1aphcollections.txt"
        self.save_edges = []
        self.save_nodes = []
        self.load_edges = []
        self.load_nodes = []
        
        self.N = 100    # Number of nodes
        self.I = 5  # Number of graph indices.
        for graph_index in xrange(self.I):
            d = np.random.random((self.N, self.N))
            g = nx.Graph()
            for i in xrange(self.N):
                for j in xrange(i+1, self.N):
                    if d[i, j] > 0.7:
                        g.add_edge(i, j)
            self.collection.graphs[graph_index] = g

        self.test_graph = nx.Graph()

    def test_setitem(self):
        """
        Test of tethne.data.GraphCollection.__setitem__ .
        """
        
        # If value is not a tuple, then it must be a Graph.
        with self.assertRaises(ValueError):
            self.collection[6] = 'asdf'
    
        # If value is a tuple, then value[0] must be a Graph.
        with self.assertRaises(ValueError):
            self.collection[6] = ('asdf', 'asdf')

        pass

    def test_save(self):
        """
        Test of tethne.data.GraphCollection.save .
        """
        
        C = ds.GraphCollection()
        C[0] = nx.Graph()
        C[0].add_edge(0,1)
        C[0].add_edge(2,1)
        C[0].add_edge(2,0)
        C.save(self.correct_filename)

        self.assertIsInstance(C,type(C), msg="yes it is a object")
        
        self.g1 = ds.GraphCollection()
        self.g2 = ds.GraphCollection()
       

        filepath = "../testsuite/testin/coauthors.txt"
        wos_list = rd.wos.parse(filepath)
        meta_list= rd.wos.convert(wos_list)


        coauthors = nt.authors.coauthors(meta_list, 'date','jtitle','ayjid')
        author_coup = nt.papers.author_coupling(meta_list, 1, 'ayjid', 'atitle')


        self.g1.__setitem__('ayjid',coauthors)
        self.g2.__setitem__('ayjid',author_coup)


        self.save_nodes = self.g1.nodes()
        self.save_edges = self.g1.edges()

        self.g1.save\
            ("../testsuite/testout/graphcollections.txt")

        obtained = set(self.g1.nodes())
        expected = set (['TUERHONG G', 'CHEN T', 'BARRENECHEA E', 'KANG P', \
                         'GALAR M', 'ZHANG B', 'CHO S', 'CHEN F', 'YE G', \
                         'MAALEJ W', 'FERNANDEZ A', 'HUANG J', 'PAGANO D', \
                         'HERMAN G', 'KIM S', 'VATALIS K', 'WANG Y', \
                         'HERRERA F', 'MODIS K'])
                
        self.assertSetEqual(obtained, expected, "nodes not as expected")
    

    
    

    def test_load(self):
        """
        Test of tethne.data.GraphCollection.load .
        """
        self.g3 = ds.GraphCollection()
       
        # Check for IOError.
        
#        with self.assertRaises('IOError',self.collection.load(self.wrong_filename)):
#            D = dt.GraphCollection()
#            D.load(self.filename)

        self.g3.load("../testsuite/testout/graphcollections.txt")
        
        self.load_nodes = set(self.g3.nodes())
        self.load_edges = self.g3.edges()
        
        obtained = set(self.g3.nodes())
        expected = set (['TUERHONG G', 'CHEN T', 'BARRENECHEA E', 'KANG P', \
                         'GALAR M', 'ZHANG B', 'CHO S', 'CHEN F', 'YE G', \
                         'MAALEJ W', 'FERNANDEZ A', 'HUANG J', 'PAGANO D', \
                         'HERMAN G', 'KIM S', 'VATALIS K', 'WANG Y', \
                         'HERRERA F', 'MODIS K'])
    
        self.assertSetEqual \
            (obtained, self.load_nodes, "nodes not as expected")
            
            
    
    def tearDown(self):
        pass




if __name__ == '__main__':
    unittest.main()



"""
Methods for analyzing :class:`.GraphCollection` objects.

For the most part, these methods simply provide systematic access to algorithms
in NetworkX.
"""

import networkx as nx
import types
import graph

def algorithm(C, method, **kwargs):
    """
    Apply NetworkX method to each ``Graph`` in :class:`.GraphCollection`\.

    Passes kwargs to specified NetworkX method for each Graph, and returns
    a dictionary of results indexed by element (node or edge) and graph index
    (e.g. ``date``).

    Parameters
    ----------
    C : :class:`.GraphCollection`
        The :class:`.GraphCollection` to analyze. The specified method will be
        applied to each :class:`.Graph` in **C**.
    method : string
        Name of a method in NetworkX to execute on graph collection.
    **kwargs
        A list of keyword arguments that should correspond to the parameters
        of the specified method.

    Returns
    -------
    results : dict
        A nested dictionary of results: results/elem(node or edge)/graph
        index.

    Raises
    ------
    ValueError
        If name is not in networkx, or if no such method exists.

    Examples
    --------

    *Betweenness centrality:*

    .. code-block:: python

       >>> import tethne.analyze as az
       >>> BC = az.collection.algorithm(C, 'betweenness_centrality')
       >>> print BC[0]
       {1999: 0.010101651117889644,
       2000: 0.0008689093723107329,
       2001: 0.010504898852426189,
       2002: 0.009338654511194512,
       2003: 0.007519105636349891}

    """

    results = {}

    if not method in nx.__dict__:
        raise(ValueError("No such name in networkx."))
    else:
        if type(nx.__dict__[method]) is not types.FunctionType:
            raise(ValueError("No such method in networkx."))
        else:
            for k, G in C.graphs.iteritems():
                r = nx.__dict__[method](G, **kwargs)
                for elem, value in r.iteritems():
                    try:
                        results[elem][k] = value
                    except KeyError:
                        results[elem] = { k: value }
                nx.set_node_attributes(G, method, r)    # [#61510128]
    return results

def connected(C, method, **kwargs):
    """
    Performs analysis methods from networkx.connected on each graph in the
    collection.

    Parameters
    ----------
    C : :class:`.GraphCollection`
        The :class:`.GraphCollection` to analyze. The specified method will be
        applied to each :class:`.Graph` in **C**.
    method : string
        Name of method in networkx.connected.
    **kwargs : kwargs
        Keyword arguments, passed directly to method.

    Returns
    -------
    results : dictionary
        Keys are graph indices, values are output of method for that graph.

    Raises
    ------
    ValueError
        If name is not in networkx.connected, or if no such method exists.

    Examples
    --------

    .. code-block:: python

        >>> import tethne.data as ds
        >>> import tethne.analyze as az
        >>> import networkx as nx
        >>> C = ds.GraphCollection()
        >>> # Generate some random graphs
        >>> for graph_index in xrange(1999, 2004):
        >>>     g = nx.random_regular_graph(4, 100)
        >>>     C[graph_index] = g
        >>> results = az.collection.connected(C, 'connected', k=None)
        >>> print results
        {1999: False,
        2000: False,
        2001: False,
        2002: False,
        2003: False }

    """

    results = {}

    if not method in nx.connected.__dict__:
        raise(ValueError("No such name in networkx.connected."))
    else:
        if type(nx.connected.__dict__[method]) is not types.FunctionType:
            raise(ValueError("No such method in networkx.connected."))
        else:
            for k, G in C.graphs.iteritems():
                results[k] = nx.connected.__dict__[method](G, **kwargs)
    return results

def node_history(C, node, attribute, verbose=False):
    """
    Returns a dictionary of attribute values for each Graph in C for a single
    node.

    Parameters
    ----------
    C : :class:`.GraphCollection`
    node : str
        The node of interest.
    attribute : str
        The attribute of interest; e.g. 'betweenness_centrality'
    verbose : bool
        If True, prints status and debug messages.

    Returns
    -------
    history : dict
        Keys are Graph keys in C; values are attribute values for node.
    """

    history = {}

    for k,G in C.graphs.iteritems():
        asdict = { v[0]:v[1] for v in G.nodes(data=True) }
        try:
            history[k] = asdict[node][attribute]
        except KeyError:
            if verbose: print "No such node or attribute in Graph " + str(k)

    return history

def edge_history(C, source, target, attribute, verbose=False):
    """
    Returns a dictionary of attribute vales for each Graph in C for a single
    edge.

    Parameters
    ----------
    C : :class:`.GraphCollection`
    source : str
        Identifier for source node.
    target : str
        Identifier for target node.
    attribute : str
        The attribute of interest; e.g. 'betweenness_centrality'
    verbose : bool
        If True, prints status and debug messages.

    Returns
    -------
    history : dict
        Keys are Graph keys in C; values are attribute values for edge.
    """

    history = {}

    for k,G in C.graphs.iteritems():
        try:
            attributes = G[source][target]
            try:
                history[k] = attributes[attribute]
            except KeyError:
                if verbose: print "No such attribute for edge in Graph " \
                                    + str(k)
        except KeyError:
            if verbose: print "No such edge in Graph " + str(k)

    return history

def node_global_closeness_centrality(C, node):
    """
    Calculates global closeness centrality for node in each graph in
    :class:`.GraphCollection` C.

    """

    results = {}
    for key, g in C.graphs.iteritems():
        results[key] = graph.node_global_closeness_centrality(g, node)

    return results


"""
Methods for parsing bibliographic datasets.

.. autosummary::

   dfr
   mallet
   pubmed
   wos

Each file reader provides methods to parse bibliographic data from a
scholarly database (e.g. Web of Science or PubMed), resulting in a
list of :class:`.Paper` instances containing as many as possible of
the following keys (missing values are set to None):

===========    =====    ===================================================
Field          Type     Description
===========    =====    ===================================================
aulast         list     Authors' surnames, as a list.
auinit         list     Authors' initials, as a list.
institution    dict     Institutions with which the authors are affiliated.
atitle         str      Article title.
jtitle         str      Journal title or abbreviated title.
volume         str      Journal volume number.
issue          str      Journal issue number.
spage          str      Starting page of article in journal.
epage          str      Ending page of article in journal.
date           int      Date of publication.
abstract       str
===========    =====    ===================================================

These keys are associated with the meta data entries in the databases of
organizations such as the International DOI Foundation and its Registration
Agencies such as CrossRef and DataCite.

In addition, :class:`.Paper` instances will contain keys with information
relevant to the networks of interest for Tethne including:

===========    =====    ======================================================
Field          Type     Description
===========    =====    ======================================================
citations      list     List of minimum :class:`.Paper` instances for cited
                        references.
ayjid          str      First author's name (last, fi), publication year, and
                        journal.
doi            str      Digital Object Identifier.
pmid           str      PubMed ID.
wosid          str      Web of Science UT fieldtag.
===========    =====    ======================================================

Missing data here also results in the above keys being set to None.

"""

import wos
import pubmed
import dfr
import mallet
import tethne.data as dt

class DataError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

def merge(P1, P2, fields=['ayjid']):
    """
    Combines two lists (P1 and P2) of :class:`.Paper` instances into a single
    list, and attempts to merge papers with matching fields. Where there are
    conflicts, values from :class:`.Paper` in P1 will be preferred.

    Parameters
    ----------
    P1 : list
        A list of :class:`.Paper` instances.
    P2 : list
        A list of :class:`.Paper` instances.
    fields : list
        Fields used to identify matching :class:`.Paper`

    Returns
    -------
    combined : list
        A list of :class:`.Paper` instances.

    Examples
    --------

    .. code-block:: python

       >>> import tethne.readers as rd
       >>> P1 = rd.wos.read("/Path/to/data1.txt")
       >>> P2 = rd.dfr.read("/Path/to/DfR")
       >>> papers = rd.merge(P1, P2, ['ayjid'])
    """

    combined = []
    del_P1 = []
    del_P2 = []

    for x in xrange(len(P1)):
        p_1 = P1[x]
        for y in xrange(len(P2)):
            p_2 = P2[y]
            match = True
            for field in fields:
                if p_1[field] != p_2[field]:
                    match = False
                    break

            if match:   # Add values first from P2 paper, then from P1 paper.
                new_p = dt.Paper()
                for key, value in p_2.iteritems():
                    if value != '' and value != None:
                        new_p[key] = value
                for key, value in p_1.iteritems():
                    if value != '' and value != None:
                        new_p[key] = value

                del_P1.append(x)    # Flag for deletion.
                del_P2.append(y)

                combined.append(new_p)

    for x in xrange(len(P1)):
        if x not in del_P1:
            combined.append(P1[x])
    for x in xrange(len(P2)):
        if x not in del_P2:
            combined.append(P2[x])

    return combined